"""Canonical structured session checklist state and on-demand rendering."""

from __future__ import annotations

from collections import OrderedDict
import hashlib
from pathlib import Path
import uuid

from gotta.compat import UTC, datetime
from gotta.content.env import SESSION_REPO_ENV, load_state_env_at_root
from gotta.actor import SESSION_ACTOR_ENV
from gotta.projection import append_jsonl, read_jsonl_records


TODO_LOG_NAME = "todo.jsonl"
TODO_ID_PREFIX = "T"

DEFAULT_SECTION_ORDER = (
    "Status",
    "File Battery",
    "Drift / Truth Battery",
    "Mechanical / Safety Battery",
    "Actor Checklist",
    "Captured Work",
)


def todo_state_path(work_dir: Path) -> Path:
    return work_dir / "state" / TODO_LOG_NAME


def is_todo_id(value: str) -> bool:
    return (
        len(value) == 9
        and value.startswith(TODO_ID_PREFIX)
        and all(ch in "0123456789ABCDEF" for ch in value[1:])
    )


def _repo_display_name(work_dir: Path) -> str:
    state = load_state_env_at_root(work_dir)
    repo_path = str(state.get(SESSION_REPO_ENV) or "").strip()
    if not repo_path:
        return "Session"
    return Path(repo_path).name.capitalize() or "Session"


def _session_actor_id(work_dir: Path) -> str:
    state = load_state_env_at_root(work_dir)
    return str(state.get(SESSION_ACTOR_ENV) or "").strip().lower()


def _next_todo_id(used_ids: set[str]) -> str:
    while True:
        candidate = f"{TODO_ID_PREFIX}{uuid.uuid4().hex[:8].upper()}"
        if candidate not in used_ids:
            return candidate


def managed_todo_id(managed_key: str) -> str:
    digest = hashlib.sha256(managed_key.encode("utf-8")).hexdigest()[:8].upper()
    return f"{TODO_ID_PREFIX}{digest}"


def todo_events(work_dir: Path) -> list[dict[str, object]]:
    return read_jsonl_records(todo_state_path(work_dir))


def todo_items(work_dir: Path) -> list[dict[str, object]]:
    items: OrderedDict[str, dict[str, object]] = OrderedDict()
    for event in todo_events(work_dir):
        op = str(event.get("op") or "")
        item_id = str(event.get("id") or "").upper()
        if not is_todo_id(item_id):
            continue
        if op == "create":
            if item_id in items:
                continue
            items[item_id] = {
                "id": item_id,
                "section": str(event.get("section") or "Captured Work").strip()
                or "Captured Work",
                "text": str(event.get("text") or "").strip(),
                "checked": bool(event.get("checked")),
                "managed_key": str(event.get("managed_key") or "").strip(),
                "created_at": str(event.get("timestamp") or ""),
            }
            continue
        if op == "check" and item_id in items:
            items[item_id]["checked"] = bool(event.get("checked"))
    return list(items.values())


def todo_payload(
    work_dir: Path,
    *,
    status: str = "all",
    limit: int = 0,
) -> dict[str, object]:
    items = todo_items(work_dir)
    if status == "open":
        filtered = [item for item in items if not bool(item["checked"])]
    elif status == "done":
        filtered = [item for item in items if bool(item["checked"])]
    else:
        filtered = list(items)
    if limit > 0:
        filtered = filtered[:limit]
    return {
        "state_path": str(todo_state_path(work_dir)),
        "locator": "todo:session",
        "follow_command": "gotta todo",
        "entry_count": len(items),
        "open_count": sum(1 for item in items if not bool(item["checked"])),
        "done_count": sum(1 for item in items if bool(item["checked"])),
        "entries": filtered,
    }


def render_todo_entry(
    text: str,
    *,
    checked: bool,
    item_id: str,
    managed_key: str = "",
) -> str:
    lines = text.splitlines() or [text]
    marker = "x" if checked else " "
    suffix = f" <!-- gotta:id:{item_id} -->"
    if managed_key:
        suffix += f" <!-- gotta:{managed_key} -->"
    head = f"- [{marker}] {lines[0].strip()}{suffix}"
    if len(lines) == 1:
        return head
    return "\n".join([head, *[f"  {line}" if line else "  " for line in lines[1:]]])


def _render_section_items(items: list[dict[str, object]]) -> list[str]:
    lines: list[str] = []
    for item in items:
        lines.extend(
            render_todo_entry(
                str(item["text"]),
                checked=bool(item["checked"]),
                item_id=str(item["id"]),
                managed_key=str(item.get("managed_key") or ""),
            ).splitlines()
        )
    return lines


def render_todo_markdown(work_dir: Path, items: list[dict[str, object]]) -> str:
    from gotta.session.registry import (
        _actor_bind_examples,
        _actor_label,
        _default_actor_summary,
    )

    grouped: OrderedDict[str, list[dict[str, object]]] = OrderedDict()
    for section in DEFAULT_SECTION_ORDER:
        grouped[section] = []
    for item in items:
        section = str(item.get("section") or "Captured Work")
        grouped.setdefault(section, [])
        grouped[section].append(item)

    lines = [
        f"# {_repo_display_name(work_dir)} Session TODO",
        "",
        f"> Generated automatically from `state/{TODO_LOG_NAME}`.",
        "> Rendered on demand from canonical state.",
        "",
        "Prefer `gotta todo ...` for routine mutation.",
        "The structured `todo` log is canonical.",
        "",
        "## Definition Of Done",
        "",
        "- Active targets examined against live local truth.",
        "- Truth conflicts corrected, removed, or explicitly parked with rationale.",
        "- Actor findings incorporated or explicitly dispositioned.",
        "- Final verification rerun against the active target set.",
        "- Final sign-off collected from the selected team.",
        "",
        "## Status",
        "",
    ]
    lines.extend(_render_section_items(grouped["Status"]) or ["- none"])
    if grouped["File Battery"]:
        lines.extend(["", "## File Battery", ""])
        lines.extend(_render_section_items(grouped["File Battery"]))
    lines.extend(
        [
            "",
            "## Drift / Truth Battery",
            "",
        ]
    )
    if grouped["Drift / Truth Battery"]:
        lines.extend(_render_section_items(grouped["Drift / Truth Battery"]))
    else:
        lines.extend(
            [
                "- Seed placeholder. Replace with concrete truth-mismatch session state as soon as it is",
                "  discovered.",
            ]
        )
    lines.extend(
        [
            "",
            "## Mechanical / Safety Battery",
            "",
        ]
    )
    if grouped["Mechanical / Safety Battery"]:
        lines.extend(_render_section_items(grouped["Mechanical / Safety Battery"]))
    else:
        lines.extend(
            [
                "- Seed placeholder. Replace with concrete mechanical, anchor, link, and",
                "  dangerous-guidance session state as soon as it is discovered.",
            ]
        )
    lines.extend(
        [
            "",
            "## Coordination Notes",
            "",
        ]
    )
    actor_actor = _session_actor_id(work_dir)
    if actor_actor:
        lines.extend(
            [
                f"- This session is already the {_actor_label(actor_actor)} actor root for this shared concern.",
                "- `WANT.md` and `GOAL.md` here are actor-local authored charters.",
                "- Use `gotta todo`, `gotta notes`, `gotta logs`, and `gotta oops` as the live readable surfaces.",
                "- Append running notes with `gotta notes append ...` from this actor root; add `--actor <actor>` only when targeting another bound actor intentionally.",
                "- Append at least one short note before requesting completion or sign-off so shared actor visibility lands before closure.",
                f"- When materially done, record actor completion through `gotta actor complete {actor_actor}`; use `gotta actor stop {actor_actor}` for a graceful operator-directed wind-down, `gotta actor fail {actor_actor}` for actual failure, and finish durable review with `gotta actor signoff {actor_actor} --summary ...`.",
            ]
        )
    else:
        lines.extend(
            [
                f"- Choose the team intentionally. Available actors today are {_default_actor_summary()}.",
                f"- `gotta actor bind <actor...>` binds sibling actor sessions inside this shared session and creates the actor-local WANT/GOAL targets; only after bind completes can `gotta want|goal --actor <actor>` address that actor. Use {_actor_bind_examples(prefix='gotta actor bind')}.",
                "- If you choose an actor, actually consult them with `gotta actor launch <actor>`.",
                "- Actor checklist pressure begins only after a launched actor starts real evidence collection.",
            ]
        )
    if grouped["Actor Checklist"]:
        lines.extend(["", "## Actor Checklist", ""])
        lines.extend(_render_section_items(grouped["Actor Checklist"]))
    lines.extend(["", "## Captured Work", ""])
    lines.extend(_render_section_items(grouped["Captured Work"]) or ["- none yet"])

    for section, section_items in grouped.items():
        if section in DEFAULT_SECTION_ORDER or not section_items:
            continue
        lines.extend(["", f"## {section}", ""])
        lines.extend(_render_section_items(section_items))
    return "\n".join(lines) + "\n"


def create_todo_item(
    work_dir: Path,
    *,
    section: str,
    text: str,
    checked: bool = False,
    managed_key: str = "",
    item_id: str = "",
    timestamp: str | None = None,
) -> dict[str, object]:
    current_items = todo_items(work_dir)
    used_ids = {str(item["id"]).upper() for item in current_items}
    normalized_id = item_id.upper().strip() if item_id else ""
    if not is_todo_id(normalized_id) or normalized_id in used_ids:
        normalized_id = _next_todo_id(used_ids)
    payload: dict[str, object] = {
        "op": "create",
        "id": normalized_id,
        "section": section.strip() or "Captured Work",
        "text": text.strip(),
        "checked": checked,
        "managed_key": managed_key.strip(),
        "timestamp": timestamp or datetime.now(tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
    }
    append_jsonl(todo_state_path(work_dir), payload)
    return payload


def ensure_managed_todo_item(
    work_dir: Path,
    *,
    section: str,
    text: str,
    managed_key: str,
) -> dict[str, object]:
    current = todo_items(work_dir)
    for item in current:
        if str(item.get("managed_key") or "") == managed_key:
            return item
    return create_todo_item(
        work_dir,
        section=section,
        text=text,
        managed_key=managed_key,
        item_id=managed_todo_id(managed_key),
    )


def set_todo_checked(
    work_dir: Path, item_id: str, *, checked: bool
) -> dict[str, object] | None:
    normalized = item_id.strip().upper()
    if not is_todo_id(normalized):
        raise SystemExit(f"invalid TODO item id: {item_id}")
    current = {str(item["id"]).upper(): item for item in todo_items(work_dir)}
    item = current.get(normalized)
    if item is None:
        raise SystemExit(f"no TODO item matched id: {normalized}")
    if bool(item["checked"]) == checked:
        return None
    payload: dict[str, object] = {
        "op": "check",
        "id": normalized,
        "checked": checked,
        "timestamp": datetime.now(tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
    }
    append_jsonl(todo_state_path(work_dir), payload)
    updated = dict(item)
    updated["checked"] = checked
    return updated


def resolve_todo_item(work_dir: Path, *, item_id: str) -> dict[str, object]:
    normalized = item_id.strip().upper()
    if not is_todo_id(normalized):
        raise SystemExit(
            f"invalid TODO item id: {item_id}. run `gotta todo show` and pass printed ids"
        )
    for item in todo_items(work_dir):
        if str(item["id"]).upper() == normalized:
            return item
    raise SystemExit(f"no TODO item matched id: {normalized}")



from __future__ import annotations

import json
import os
from pathlib import Path
import tempfile
from typing import Any

from gotta.content.model import ContentError
from gotta.content.path import sanitize_name

PRIVATE_DIR_MODE = 0o700
PRIVATE_FILE_MODE = 0o600


def _chmod_best_effort(path: Path, mode: int) -> None:
    try:
        if path.exists() and not path.is_symlink():
            path.chmod(mode)
    except OSError:
        pass


def ensure_private_dir(path: Path) -> Path:
    path.mkdir(parents=True, exist_ok=True)
    _chmod_best_effort(path, PRIVATE_DIR_MODE)
    return path


def write_bytes_atomic(path: Path, data: bytes) -> Path:
    ensure_private_dir(path.parent)
    with tempfile.NamedTemporaryFile(dir=path.parent, delete=False) as handle:
        handle.write(data)
        handle.flush()
        os.fsync(handle.fileno())
        temp_path = Path(handle.name)
    _chmod_best_effort(temp_path, PRIVATE_FILE_MODE)
    temp_path.replace(path)
    _chmod_best_effort(path, PRIVATE_FILE_MODE)
    return path


def write_text_atomic(path: Path, text: str) -> Path:
    return write_bytes_atomic(path, text.encode("utf-8"))


def write_text_if_changed(path: Path, text: str) -> Path:
    ensure_private_dir(path.parent)
    if path.exists():
        try:
            if path.read_text(encoding="utf-8") == text:
                return path
        except OSError:
            pass
    write_text_atomic(path, text)
    return path


def append_jsonl_line(path: Path, payload: dict[str, Any]) -> None:
    ensure_private_dir(path.parent)
    fd = os.open(path, os.O_APPEND | os.O_CREAT | os.O_WRONLY, PRIVATE_FILE_MODE)
    try:
        os.write(fd, (json.dumps(payload, sort_keys=True) + "\n").encode("utf-8"))
    finally:
        os.close(fd)
    _chmod_best_effort(path, PRIVATE_FILE_MODE)


def timestamp_log_path(logs_dir: Path, timestamp: str) -> Path:
    candidate = logs_dir / timestamp
    if not candidate.exists() and not candidate.is_symlink():
        return candidate
    index = 1
    while True:
        candidate = logs_dir / f"{timestamp}--{index:02d}"
        if not candidate.exists() and not candidate.is_symlink():
            return candidate
        index += 1


def ensure_name_link(names_dir: Path, data_path: Path, preferred_name: str) -> Path:
    ensure_private_dir(names_dir)
    name_link = names_dir / sanitize_name(preferred_name)
    if name_link.exists() or name_link.is_symlink():
        if not name_link.is_symlink():
            raise ContentError(
                f"integrity error: expected symlink at {name_link}, found regular file"
            )
        target = os.readlink(name_link)
        expected = os.path.relpath(data_path, start=names_dir)
        if target != expected:
            raise ContentError(
                f"integrity error: {name_link} points to {target!r} instead of {expected!r}"
            )
        return name_link
    name_link.symlink_to(os.path.relpath(data_path, start=names_dir))
    return name_link


def read_json_object(path: Path | None) -> dict[str, Any]:
    if path is None or not path.exists():
        return {}
    try:
        payload = json.loads(path.read_text(encoding="utf-8"))
    except (OSError, json.JSONDecodeError) as exc:
        raise ContentError(f"invalid metadata file at {path}: {exc}") from exc
    if not isinstance(payload, dict):
        raise ContentError(f"invalid metadata file at {path}: expected object")
    return payload



from __future__ import annotations

import array
import hashlib
import io
import os
import select
import sys

try:
    import fcntl
    import termios
except ImportError:  # pragma: no cover - platform-specific fallback
    fcntl = None
    termios = None

from gotta.actors import resolve_actor_context
from gotta.content.env import (
    ACTOR_ID_ENV,
    CONTEXT_ID_ENV,
    CONTEXT_SOURCE_ENV,
    SESSION_ACTOR_ENV,
)
from gotta.content.model import ContextBinding
from gotta import topology


def current_context_binding() -> ContextBinding:
    explicit = os.environ.get(CONTEXT_ID_ENV, "").strip()
    if explicit:
        source = os.environ.get(CONTEXT_SOURCE_ENV, "").strip() or "explicit"
        return ContextBinding(
            context_id=explicit,
            context_source=source,
            binding_id=session_token(explicit),
        )
    codex_thread = os.environ.get("CODEX_THREAD_ID", "").strip()
    if codex_thread:
        return ContextBinding(
            context_id=codex_thread,
            context_source="codex_thread",
            binding_id=session_token(codex_thread),
        )
    term_session = os.environ.get("TERM_SESSION_ID", "").strip()
    if term_session:
        return ContextBinding(
            context_id=term_session,
            context_source="terminal_session",
            binding_id=session_token(term_session),
        )
    values = [
        os.environ.get("TTY", "").strip(),
        os.environ.get("SHELL", "").strip(),
        os.environ.get("TERM_PROGRAM", "").strip(),
        os.environ.get("TERM", "").strip(),
        os.environ.get("COPILOT_LOADER_PID", "").strip(),
        os.environ.get("COPILOT_CLI_BINARY_VERSION", "").strip(),
    ]
    if not values[0]:
        values.append(os.getcwd())
    payload = "\n".join(values)
    context_id = hashlib.sha256(payload.encode("utf-8")).hexdigest()
    return ContextBinding(
        context_id=context_id,
        context_source="terminal_fingerprint",
        binding_id=session_token(context_id),
    )


def session_token(context_id: str) -> str:
    return hashlib.sha256(context_id.encode("utf-8")).hexdigest()[:12]


def default_session_id(context_id: str) -> str:
    return session_token(context_id)


def current_actor(*, default_actor: str = "") -> str:
    fallback = (
        os.environ.get(ACTOR_ID_ENV, "").strip()
        or default_actor.strip()
        or os.environ.get(SESSION_ACTOR_ENV, "").strip()
        or current_context_binding().binding_id
    )
    speaker = resolve_actor_context(default_speaker=fallback).speaker
    normalized = topology.normalize_identity(str(speaker or fallback).strip())
    if normalized and not topology.is_placeholder_identity(normalized):
        return normalized
    fallback_normalized = topology.normalize_identity(fallback)
    if fallback_normalized and not topology.is_placeholder_identity(
        fallback_normalized
    ):
        return fallback_normalized
    return current_context_binding().binding_id


def stdin_has_meaningful_text() -> bool:
    stream = sys.stdin
    try:
        if stream.isatty():
            return False
    except Exception:
        return False
    getvalue = getattr(stream, "getvalue", None)
    if callable(getvalue):
        try:
            value = getvalue()
            if not isinstance(value, str):
                return bool(value)
            tell = getattr(stream, "tell", None)
            cursor_value = tell() if callable(tell) else 0
            cursor = cursor_value if isinstance(cursor_value, int) else 0
            return len(value) > cursor
        except Exception:
            return False
    try:
        fileno = stream.fileno()
    except (AttributeError, OSError, io.UnsupportedOperation):
        return False
    try:
        readable, _, _ = select.select([fileno], [], [], 0)
    except (OSError, ValueError):
        return False
    if not readable:
        return False
    buffer = getattr(stream, "buffer", None)
    if buffer is not None and hasattr(buffer, "peek"):
        try:
            return bool(buffer.peek(1))
        except Exception:
            pass
    if fcntl is not None and termios is not None:
        try:
            available = array.array("i", [0])
            fcntl.ioctl(fileno, termios.FIONREAD, available, True)
            return bool(available[0])
        except Exception:
            pass
    return False


def stdin_has_readable_text() -> bool:
    return stdin_has_meaningful_text()



"""Actor lifecycle status payload synthesis."""

from __future__ import annotations

import os
from pathlib import Path
import time

from gotta.actor import requested_disposition_label
from gotta.compat import datetime
from gotta.notes import actor_notes_ready, actor_voice
from gotta.session.activity import (
    _actor_evidence_note,
    _actor_evidence_summary,
    _actor_note_check_summary,
    _actor_note_summary,
    _actor_recent_activity,
)
from gotta.session.registry import (
    ACTOR_STALL_SECONDS,
    _actor_events_path,
    _actor_label,
    _actor_session_dir,
    _actor_state_path,
    _normalize_actor_name,
    _read_actor_state,
    _resolve_bound_actor_name,
)
from gotta.session.status.progress import _actor_progress_summary


ACTOR_RUNNING_STATUS = {
    "starting",
    "active",
    "closing",
    "producing_evidence",
}

ACTOR_TERMINAL_STATUS = {
    "completed",
    "failed",
    "incomplete",
    "rejected",
    "signed_off",
}


def _int_value(value: object, *, default: int = 0) -> int:
    try:
        return int(str(value or default))
    except ValueError:
        return default


def _lifecycle_entries(value: object) -> list[dict[str, object]]:
    if not isinstance(value, list):
        return []
    return [dict(item) for item in value if isinstance(item, dict)]


def _actor_status_payload(work_dir: Path, actor_name: str) -> dict[str, object]:
    actor_name = _resolve_bound_actor_name(work_dir, actor_name)
    state = _read_actor_state(work_dir, actor_name)
    status = str(state.get("status") or "pending")
    requested_status = str(state.get("requested_status") or "")
    requested_summary = str(state.get("requested_summary") or "")
    requested_label = requested_disposition_label(state)
    heartbeat_at = str(state.get("heartbeat_at") or "")
    derived_status = status
    heartbeat_stale = False
    runtime_live: bool | None = None
    try:
        pid = _int_value(state.get("pid"))
    except (TypeError, ValueError):
        pid = 0
    if pid > 0:
        try:
            os.kill(pid, 0)
        except OSError:
            runtime_live = False
        else:
            runtime_live = True
    if status in {"starting", "active"} and heartbeat_at:
        try:
            heartbeat_dt = datetime.fromisoformat(heartbeat_at.replace("Z", "+00:00"))
        except ValueError:
            heartbeat_dt = None
        if heartbeat_dt is not None:
            age = time.time() - heartbeat_dt.timestamp()
            state["heartbeat_age_seconds"] = round(age, 1)
            if age > ACTOR_STALL_SECONDS:
                derived_status = "stalled"
                heartbeat_stale = True
    elif status in {"starting", "active"} and not heartbeat_at:
        started_at = str(state.get("started_at") or "")
        try:
            started_dt = datetime.fromisoformat(started_at.replace("Z", "+00:00"))
        except ValueError:
            started_dt = None
        if (
            started_dt is not None
            and (time.time() - started_dt.timestamp()) > ACTOR_STALL_SECONDS
        ):
            derived_status = "stalled"
            heartbeat_stale = True
    signoff_at = str(state.get("signoff_at") or "")
    actor_dir = _actor_session_dir(work_dir, actor_name)
    voice = actor_voice(work_dir, _normalize_actor_name(actor_name))
    notes_ready = voice == "present"
    evidence = _actor_evidence_summary(work_dir, actor_name)
    evidence_note = _actor_evidence_note(evidence)
    recent_activity = _actor_recent_activity(work_dir, actor_name)
    note_summary = _actor_note_summary(work_dir, actor_name)
    note_check_summary = _actor_note_check_summary(work_dir, actor_name)
    progress = _actor_progress_summary(work_dir, actor_name)
    lifecycle_entries = _lifecycle_entries(recent_activity.get("recent_lifecycle"))
    if (
        lifecycle_entries
        and str(lifecycle_entries[0].get("event") or "") == "runtime_exit"
    ):
        lifecycle_detail = str(lifecycle_entries[0].get("detail") or "")
        request_labels = {
            "stop_requested": "graceful stop request",
            "failed_requested": "failure request",
            "signoff_requested": "sign-off request",
            "complete_requested": "completion request",
        }
        honored = next(
            (
                request_labels.get(str(item.get("event") or ""))
                for item in lifecycle_entries[1:]
                if str(item.get("event") or "") in request_labels
            ),
            "",
        )
        if honored and "code 0" in lifecycle_detail:
            lifecycle_entries[0]["summary"] = (
                f"runtime exit: actor process exited cleanly after honoring {honored}"
            )
    if lifecycle_entries:
        recent_activity["recent_lifecycle"] = lifecycle_entries
        recent_activity["last_lifecycle_at"] = str(
            lifecycle_entries[0].get("timestamp") or ""
        )
        recent_activity["last_lifecycle_summary"] = str(
            lifecycle_entries[0].get("summary") or ""
        )
    evidence_live = _int_value(evidence.get("artifact_count")) > 0
    if signoff_at:
        derived_status = "signed_off"
    notes_status = (
        "present"
        if actor_notes_ready(work_dir, _normalize_actor_name(actor_name))
        else "empty"
    )
    runtime_note = ""
    if status in {"starting", "active"} and runtime_live is False:
        if requested_status in {"completed", "failed", "signed_off"}:
            derived_status = requested_status
        else:
            derived_status = "awaiting_disposition"
        runtime_note = (
            " Actor runtime is no longer live, so this is awaiting an explicit "
            "completion, failure, or sign-off record."
        )
    if derived_status in {"starting", "active"} and evidence_live:
        derived_status = "producing_evidence"
    if requested_status and derived_status in {
        "starting",
        "active",
        "producing_evidence",
    }:
        derived_status = "closing"
    still_running = derived_status in ACTOR_RUNNING_STATUS and not heartbeat_stale
    request_note = ""
    if requested_status and derived_status not in ACTOR_TERMINAL_STATUS:
        if requested_label == "stop":
            if derived_status == "stalled":
                request_note = (
                    " Operator already requested a graceful stop"
                    + (f" ({requested_summary})." if requested_summary else ".")
                    + " Because the heartbeat is stale, you can settle now with "
                    f"`gotta actor settle {_normalize_actor_name(actor_name)}`."
                )
            else:
                request_note = (
                    " Operator already requested a graceful stop"
                    + (f" ({requested_summary})." if requested_summary else ".")
                    + " The actor should wind down, append one final short note, and sign off."
                )
        elif derived_status == "stalled":
            request_note = (
                f" Operator already requested `{requested_label}`"
                + (f" ({requested_summary})." if requested_summary else ".")
                + " Because the heartbeat is stale, you can settle now with "
                f"`gotta actor settle {_normalize_actor_name(actor_name)}`."
            )
        else:
            request_note = (
                f" Operator already requested `{requested_label}`"
                + (f" ({requested_summary})." if requested_summary else ".")
                + " That pending disposition will become authoritative automatically "
                "when the actor runtime exits."
            )
    voice_missing = voice == "missing"
    voice_setup = voice == "setup"
    voice_pulse = voice == "pulse"
    progress_stale = bool(progress.get("progress_stale"))
    last_note_at = str(note_summary.get("last_note_at") or "")
    last_artifact_at = str(evidence.get("last_artifact_at") or "")
    note_checks_since_update = _int_value(
        note_check_summary.get("note_checks_since_update")
    )
    needs_note_refresh = bool(
        evidence_live
        and last_artifact_at
        and (not last_note_at or last_artifact_at > last_note_at)
    )
    low_signal_progress = (
        bool(runtime_live)
        and progress_stale
        and _int_value(evidence.get("artifact_count")) == 0
    )
    if derived_status == "closing":
        if notes_ready and needs_note_refresh:
            next_step = (
                "actor close-out is pending while the runtime is still live, and new "
                "actor-attributed evidence landed after the last short note. "
                + (evidence_note + " " if evidence_note else "")
                + "Land one final short note now so the close-out reflects the latest evidence, "
                "then wait for runtime exit before treating the terminal disposition as authoritative."
                + request_note
                + runtime_note
            )
        elif notes_ready:
            next_step = (
                "actor close-out is pending while the runtime is still live. "
                + (evidence_note + " " if evidence_note else "")
                + "Let the actor finish the current wave, then wait for runtime exit before "
                "treating the terminal disposition as authoritative."
                + request_note
                + runtime_note
            )
        elif voice_pulse:
            next_step = (
                "actor close-out is pending while the runtime is still live. Non-note signal is "
                "already landing through friction or shared evidence, but the final short actor "
                "note is still missing. "
                + (evidence_note + " " if evidence_note else "")
                + "Land one final short actor-authored note before runtime exit so the close-out "
                "has durable voice, then recheck actor status."
                + request_note
                + runtime_note
            )
        elif voice_setup:
            next_step = (
                "actor close-out is pending while the runtime is still live. Setup note is present, "
                "but actor voice is still missing. "
                + (evidence_note + " " if evidence_note else "")
                + "Land one final short actor-authored note before runtime exit so the close-out "
                "has real voice, then recheck actor status."
                + request_note
                + runtime_note
            )
        else:
            next_step = (
                "actor close-out is pending while the runtime is still live, but actor voice is "
                "still missing. "
                + (evidence_note + " " if evidence_note else "")
                + "Land one final short actor-authored note before runtime exit so the close-out "
                "has voice, then recheck actor status." + request_note + runtime_note
            )
    elif derived_status == "producing_evidence":
        if notes_ready and needs_note_refresh:
            next_step = (
                "actor is still active and producing evidence artifacts, and new evidence landed "
                "after the last short note. "
                + (evidence_note + " " if evidence_note else "")
                + "Land a short note now so the latest evidence wave has durable actor narration, "
                f"then recheck `gotta actor status {_normalize_actor_name(actor_name)}` shortly."
                + request_note
                + runtime_note
            )
        elif notes_ready:
            next_step = (
                "actor is still active and producing evidence artifacts. "
                + (evidence_note + " " if evidence_note else "")
                + "Use `gotta notes` for live actor visibility; recheck `gotta actor status "
                f"{_normalize_actor_name(actor_name)}` shortly before closing the actor out."
                + request_note
                + runtime_note
            )
        elif voice_pulse:
            next_step = (
                "actor is still active and producing evidence artifacts. Non-note signal is "
                "already present through friction or shared evidence, but the first short actor "
                "note has not landed yet. "
                + (evidence_note + " " if evidence_note else "")
                + "Let the current evidence wave finish, then append a short actor-authored note "
                "before requesting completion or sign-off."
                + request_note
                + runtime_note
            )
        elif voice_setup:
            next_step = (
                "actor is still active and producing evidence artifacts. Setup note is present, "
                "but actor voice has not landed yet. "
                + (evidence_note + " " if evidence_note else "")
                + "Append a short actor-authored note before requesting completion or sign-off, "
                f"then recheck `gotta actor status {_normalize_actor_name(actor_name)}` shortly."
                + request_note
                + runtime_note
            )
        else:
            next_step = (
                "actor is still active and producing evidence artifacts, but actor voice is still "
                "missing. "
                + (evidence_note + " " if evidence_note else "")
                + "Append a short actor-authored note before requesting completion or sign-off, "
                f"then recheck `gotta actor status {_normalize_actor_name(actor_name)}` shortly."
                + request_note
                + runtime_note
            )
    elif (
        derived_status in {"starting", "active"} and notes_ready and needs_note_refresh
    ):
        next_step = (
            "actor is still active and new actor-attributed evidence landed after the last short "
            "note. "
            + (evidence_note + " " if evidence_note else "")
            + "Land a short note now so the current evidence wave is narrated before close-out."
            + request_note
            + runtime_note
        )
    elif derived_status in {"starting", "active"} and notes_ready:
        next_step = (
            "actor is still active and actor voice is present. "
            "Use `gotta notes` for live actor visibility; recheck `gotta actor status "
            f"{_normalize_actor_name(actor_name)}` shortly before closing the actor out."
            + request_note
            + runtime_note
        )
    elif derived_status in {"starting", "active"} and voice_pulse:
        next_step = (
            "actor is live and non-note signal is already landing through friction or shared "
            "evidence, but the first short actor note has not landed yet. Give the runtime a brief "
            "window to turn that signal into a short note before treating this as a "
            "visibility failure." + request_note + runtime_note
        )
    elif derived_status in {"starting", "active"} and voice_setup:
        next_step = (
            "setup note is present, but actor voice is still missing. Give the runtime a brief "
            "startup window to land the first short actor-authored note before treating this as a "
            "visibility failure. If actor voice is still missing after one heartbeat interval or "
            "after the first materialized artifact, intervene and recheck actor status."
            + request_note
            + runtime_note
        )
    elif derived_status in {"starting", "active"}:
        next_step = (
            "actor is live, but actor voice is still missing. Give the runtime a brief startup "
            "window to land the first short actor-authored note before treating this as a "
            "visibility failure. If actor voice is still missing after one heartbeat interval or "
            "after the first materialized artifact, intervene and recheck actor status."
            + request_note
            + runtime_note
        )
    elif derived_status == "awaiting_disposition":
        next_step = (
            "actor runtime is no longer running, but no durable terminal lifecycle was recorded yet. "
            + (evidence_note + " " if evidence_note else "")
            + "Inspect `gotta notes` plus the shared evidence web, then settle with "
            f"`gotta actor settle {_normalize_actor_name(actor_name)}`"
            + (
                f" to honor the pending `{requested_label}` request."
                if requested_status
                else " to record the authoritative terminal disposition."
            )
            + request_note
        )
    elif derived_status == "stalled" and (not voice_missing or evidence_live):
        next_step = (
            "actor heartbeat is stale, but material actor state already exists in `gotta notes` or the "
            "shared evidence web. "
            + (evidence_note + " " if evidence_note else "")
            + "Inspect the notes and decide whether to wait, relaunch, or disposition manually."
            + request_note
        )
    elif derived_status == "completed" and (notes_ready or evidence_live):
        if notes_ready and needs_note_refresh:
            next_step = (
                "actor run is complete, but new actor-attributed evidence landed after the last "
                "short note. Land one short note now, then record durable sign-off intentionally."
            )
        elif notes_ready:
            next_step = (
                "actor run is complete; inspect `gotta notes` plus the shared evidence web, then record "
                "durable sign-off with "
                f"`gotta actor signoff {_normalize_actor_name(actor_name)} --summary ...`."
            )
        elif voice_pulse:
            next_step = (
                "actor run is complete and non-note signal landed through friction or shared "
                "evidence, but the final short note is still missing. Add one short actor-authored "
                "note now, then sign off intentionally."
            )
        else:
            next_step = (
                "actor run is complete and evidence landed, but actor voice is still missing. Wait "
                "for a short actor-authored note or sign off intentionally only if you are "
                "explicitly accepting an evidence-only actor contribution."
            )
    elif derived_status == "incomplete":
        next_step = (
            "actor finished without material notes or evidence. Decide whether to relaunch, "
            "fail, or sign off intentionally."
        )
    elif derived_status == "failed" and evidence_note:
        next_step = (
            "actor was manually marked failed, but evidence already landed in shared state. "
            + evidence_note
            + " Keep or reject that evidence intentionally instead of assuming it vanished."
        )
    elif (
        derived_status in {"pending", "bound"}
        and notes_ready
        and evidence_live
        and needs_note_refresh
    ):
        next_step = (
            "actor already has actor-authored narration and shared evidence, but new evidence "
            "landed after the last short note. "
            + (evidence_note + " " if evidence_note else "")
            + "Land a short note now, then keep landing short notes after each substantive "
            "evidence wave so review, handoff, and session-wide inspection surfaces stay current."
        )
    elif derived_status in {"pending", "bound"} and notes_ready and evidence_live:
        next_step = (
            "actor already has actor-authored narration and shared evidence without an active runtime. "
            + (evidence_note + " " if evidence_note else "")
            + "Keep landing notes as the session evolves; when this actor's contribution is "
            "materially complete, record the authoritative close-out intentionally with "
            f"`gotta actor signoff {_normalize_actor_name(actor_name)} --summary ...`."
        )
    elif derived_status in {"pending", "bound"} and voice_pulse:
        next_step = (
            "non-note signal is present through friction or shared evidence, but no short actor "
            "note has landed yet. "
            + (evidence_note + " " if evidence_note else "")
            + "Land one short actor-authored note now, then keep landing short notes after each "
            "material evidence wave so review, handoff, and session-wide inspection surfaces have "
            "continuous actor voice."
        )
    elif derived_status in {"pending", "bound"} and notes_ready:
        next_step = (
            "actor already has actor-authored narration but no shared evidence artifacts yet. "
            "Continue retrieval if more evidence should land, or close out intentionally once "
            f"the narrative is complete with `gotta actor signoff {_normalize_actor_name(actor_name)} --summary ...`."
        )
    elif derived_status in {"pending", "bound"} and voice_setup:
        next_step = (
            "setup note is present, but actor voice has not landed yet. Continue retrieval until the "
            "actor writes a short note, or close this branch out intentionally only if setup-only "
            "state is truly sufficient."
        )
    elif evidence_live and not notes_ready:
        next_step = (
            "actor-attributed evidence is already live in the shared session web, but actor voice is "
            "still missing. "
            + (evidence_note + " " if evidence_note else "")
            + "Land one short actor-authored note now, then keep landing short notes as the story "
            "moves so review, handoff, and session-wide inspection surfaces have actor voice "
            "instead of evidence-only state."
        )
    else:
        next_step = ""
    pulse_next_step = ""
    if note_checks_since_update > 0 and derived_status not in ACTOR_TERMINAL_STATUS:
        check_noun = "time" if note_checks_since_update == 1 else "times"
        if last_note_at:
            pulse_next_step = (
                f"Supervisor has checked this actor's notes {note_checks_since_update} {check_noun} "
                "since the last note. Land one short note now if real progress exists."
            )
        else:
            pulse_next_step = (
                f"Supervisor has checked this actor's notes {note_checks_since_update} {check_noun} "
                "and no first short note has landed yet. Land one short note now if real progress exists."
            )
    if low_signal_progress:
        next_step = (
            "actor runtime is still live, but actor-authored progress is stale and no "
            "actor-attributed evidence has landed yet. Treat this as a low-signal run until "
            "fresh actor-authored progress or evidence appears."
            + request_note
            + runtime_note
        )
    if pulse_next_step:
        next_step = f"{pulse_next_step} {next_step}".strip()
    return {
        **state,
        "label": _actor_label(actor_name, work_dir=work_dir),
        "status": derived_status,
        "state_path": str(_actor_state_path(work_dir, actor_name)),
        "events_path": str(_actor_events_path(work_dir, actor_name)),
        "actor_dir": str(actor_dir),
        "notes_status": notes_status,
        "notes_ready": notes_ready,
        "voice": voice,
        "evidence_live": bool(evidence_live),
        "evidence_note": evidence_note,
        "requested_status": requested_status,
        "requested_summary": requested_summary,
        "requested_label": requested_label,
        "requested_pending": bool(
            requested_status and derived_status not in ACTOR_TERMINAL_STATUS
        ),
        "still_running": still_running,
        "runtime_live": runtime_live,
        "review_ready": bool(
            derived_status in {"completed", "signed_off"}
            and (notes_ready or evidence_live)
        ),
        "next_step": next_step,
        **note_summary,
        **note_check_summary,
        **progress,
        **recent_activity,
        **evidence,
    }



"""Artifact materialization for dispatch."""

from __future__ import annotations

import os
from typing import Any

from gotta.actor import require_writer, session_actor, writer_name
from gotta.actors import ACTOR_SPEAKER_ENV
from gotta.capture import Capture
from gotta.content.env import ACTOR_ID_ENV
from gotta.content.model import (
    CommonOptions,
    ContentError,
    Materialization,
    ResolvedDirs,
)
from gotta.content.scope import session_identity
from gotta.content.store import materialize_bytes
from gotta.dispatch.metadata import _derived_source_metadata
from gotta.resolve.intent import (
    SUPPRESS_MATERIALIZATION_ENV as INVOCATION_SUPPRESS_MATERIALIZATION_ENV,
)
from gotta.resolve.invoke import (
    ResolvedInvocation,
    invocation_locator,
    resolve_invocation,
)


SUPPRESS_MATERIALIZATION_ENV = INVOCATION_SUPPRESS_MATERIALIZATION_ENV


def _materialize_invocation(
    resolved_or_plugin: ResolvedInvocation | str,
    argv_or_data: list[str] | bytes,
    options: CommonOptions | None = None,
    data: bytes | None = None,
    capture: Capture | None = None,
    *,
    dirs: ResolvedDirs,
) -> Materialization | None:
    if isinstance(resolved_or_plugin, ResolvedInvocation):
        resolved = resolved_or_plugin
        payload = argv_or_data if isinstance(argv_or_data, bytes) else b""
    else:
        plugin = resolved_or_plugin
        argv = list(argv_or_data) if isinstance(argv_or_data, list) else []
        payload = data or b""
        resolved = resolve_invocation(plugin, argv, options or CommonOptions())
    if not payload:
        return None
    materialize_plugin = resolved.resolved_plugin
    materialize_argv = resolved.resolved_argv
    explicit_actor = os.environ.get(ACTOR_ID_ENV, "").strip()
    actor = explicit_actor or session_identity(dirs.session_dir)
    target_actor = explicit_actor or session_actor(dirs.session_dir) or actor
    resolved_session_root = dirs.session_dir.resolve()
    actor_branch = resolved_session_root.parent.name == "actors"
    if target_actor and actor_branch:
        writer = writer_name()
        if explicit_actor and not os.environ.get(ACTOR_SPEAKER_ENV, "").strip():
            writer = explicit_actor
        try:
            require_writer(
                dirs.session_dir,
                target_actor,
                writer=writer,
                action="attribute materialized artifacts to this actor branch",
            )
        except SystemExit as exc:
            raise ContentError(str(exc)) from exc
    metadata: dict[str, Any] = {
        "tool": "gotta",
        "plugin": materialize_plugin,
        "provider": resolved.provider,
        "artifact_kind": resolved.artifact_kind,
        "subcommand": materialize_argv[0] if materialize_argv else "",
        "argv": materialize_argv,
        "locator": invocation_locator(materialize_plugin, materialize_argv),
        "canonical_locator": resolved.canonical_locator,
        "source_kind": "stdin"
        if resolved.entry_plugin == "read" and resolved.entry_argv == ["-"]
        else "render",
        "content_type": capture.type
        if capture is not None and capture.type
        else resolved.content_type,
        "session_dir": str(dirs.session_dir),
        "content_dir": str(dirs.content_dir),
        "actor": actor,
    }
    if resolved.entry_plugin != materialize_plugin:
        metadata["entrypoint"] = resolved.entry_plugin
        metadata["entry_argv"] = resolved.entry_argv
        metadata["entry_locator"] = invocation_locator(
            resolved.entry_plugin, resolved.entry_argv
        )
        metadata["provider"] = resolved.provider
    actor_dir = os.environ.get("GOTTA_ACTOR_DIR", "").strip()
    if actor_dir:
        metadata["actor_dir"] = actor_dir
    invocation_id = os.environ.get("GOTTA_INVOCATION_ID", "").strip()
    if invocation_id:
        metadata["invocation_id"] = invocation_id
    metadata.update(
        _derived_source_metadata(
            materialize_plugin,
            materialize_argv,
            payload,
            provider=resolved.provider,
        )
    )
    if capture is not None:
        metadata.update(capture.meta)
    preferred_name = resolved.preferred_name
    if capture is not None and capture.name and not (options and options.save_as):
        preferred_name = capture.name
    return materialize_bytes(
        payload,
        dirs=dirs,
        preferred_name=preferred_name,
        metadata=metadata,
    )



"""Top-level session-rooted procedural trace surface."""

from __future__ import annotations

import argparse
import json

from gotta.actor import require_writer, session_actor, writer_name, writer_role
from gotta.content.scope import session_is_initialized
from gotta.helptext import format_long_help, is_long_help_request
from gotta.logs import append_log_record, log_records, logs_state_path
from gotta.session import activity as session_activity
from gotta.session import charter as session_charter
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope
from gotta import topology


def _add_root_args(parser: argparse.ArgumentParser) -> None:
    session_charter.add_target_args(parser)


def build_parser(command_name: str = "gotta logs") -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=(
            "Inspect or mutate the canonical session procedural trace. "
            "Shared-session read paths show all bound actor logs by default. "
            "Use notes for actor-authored narration; use logs for chronology and system/runtime trace."
        ),
    )
    parser.add_argument(
        "action",
        nargs="?",
        choices=["show", "append", "extend"],
        help="show entries, append one entry, or extend with many entries",
    )
    parser.add_argument(
        "value",
        nargs="*",
        help="inline log text for append/extend",
    )
    _add_root_args(parser)
    parser.add_argument(
        "--from-file",
        help="read log text from a UTF-8 file instead of inline text; use '-' for stdin",
    )
    parser.add_argument(
        "--stdin",
        dest="use_stdin",
        action="store_true",
        help="read log text from stdin explicitly",
    )
    parser.add_argument("--output", choices=["json", "text"], default="text")
    parser.add_argument("--limit", type=int, default=0)
    return parser


def session_access_mode(argv: list[str]) -> str:
    positionals = session_charter.argv_positionals(
        argv,
        valued_flags=(
            "--session",
            "--actor",
            "--from-file",
            "--output",
            "--limit",
        ),
    )
    action = positionals[0] if positionals else "show"
    return "write" if action in {"append", "extend"} else "read"


def _aggregate_logs(
    work_dir,
    *,
    actor_name: str | None = None,
) -> tuple[list[str], list[dict[str, object]]]:
    actor_ids = list(session_scope._target_actor_ids(work_dir, actor_name))
    records: list[dict[str, object]] = []
    for current_actor in actor_ids:
        actor_root = session_registry._actor_session_dir(work_dir, current_actor)
        for record in log_records(actor_root):
            if (
                actor_root.resolve().parent.name == "actors"
                and writer_role(
                    actor_root, current_actor, writer=str(record.get("actor") or "")
                )
                == "foreign"
            ):
                continue
            payload = dict(record)
            payload.setdefault("actor", current_actor)
            payload.setdefault(
                "label", session_registry._actor_label(current_actor, work_dir=work_dir)
            )
            records.append(payload)
    records = sorted(records, key=lambda item: str(item.get("timestamp") or ""))
    return actor_ids, records


def _is_exact_session_root(work_dir) -> bool:
    resolved = work_dir.resolve()
    return (
        resolved.parent.name != "actors"
        and topology.parse_grouped_session_root(resolved) is None
        and topology.parse_shared_session_root(resolved) is None
        and session_is_initialized(resolved)
    )


def main(argv: list[str] | None = None) -> int:
    argv = list(argv or [])
    if is_long_help_request(argv):
        print(format_long_help(build_parser()))
        return 0
    parser = build_parser()
    try:
        args = parser.parse_args(argv)
    except SystemExit as exc:
        if int(exc.code or 0) == 0:
            return 0
        raise
    action = args.action or "show"
    if action == "show":
        work_dir, scoped_actor = session_scope._observation_scope(
            explicit_session=getattr(args, "session", None),
            explicit_actor=getattr(args, "actor", None),
        )
        if scoped_actor:
            visible = [
                record
                for record in log_records(work_dir)
                if writer_role(
                    work_dir, scoped_actor, writer=str(record.get("actor") or "")
                )
                != "foreign"
            ]
            entries = sorted(visible, key=lambda item: str(item.get("timestamp") or ""))
            limited = (
                entries[-max(args.limit, 0) :] if max(args.limit, 0) > 0 else entries
            )
            locator = session_charter._native_surface_locator(
                "logs", actor_name=scoped_actor
            )
            payload = {
                "state_path": str(logs_state_path(work_dir)),
                "locator": locator,
                "follow_command": session_charter._native_surface_follow_command(
                    "logs",
                    actor_name=scoped_actor,
                ),
                "entry_count": len(entries),
                "entries": limited,
            }
            if args.output == "json":
                print(json.dumps(payload, indent=2, sort_keys=True))
                return 0
            print(f"logs: {payload['follow_command']}")
            print(f"entries: {payload['entry_count']}")
            for record in payload["entries"]:
                timestamp = str(record.get("timestamp") or "unknown-time")
                actor = str(record.get("actor") or scoped_actor)
                message = (
                    str(record.get("message") or "").strip() or "unspecified log entry"
                )
                message_lines = message.splitlines() or ["unspecified log entry"]
                print(f"- `{timestamp}` [{actor}] {message_lines[0]}")
                for continuation in message_lines[1:]:
                    print(f"  {continuation}")
            return 0
        actor_ids, records = _aggregate_logs(work_dir)
        if not actor_ids:
            if not _is_exact_session_root(work_dir):
                raise SystemExit(
                    "no actors bound for this session; bind one intentionally with "
                    + session_registry._actor_bind_examples(prefix="gotta actor bind")
                )
            entries = sorted(
                log_records(work_dir), key=lambda item: str(item.get("timestamp") or "")
            )
            limited = (
                entries[-max(args.limit, 0) :] if max(args.limit, 0) > 0 else entries
            )
            payload = {
                "session_root": str(work_dir),
                "actor_count": 0,
                "actors": [],
                "entry_count": len(entries),
                "entries": limited,
                "state_path": str(logs_state_path(work_dir)),
                "locator": session_charter._native_surface_locator("logs"),
                "follow_command": session_charter._native_surface_follow_command(
                    "logs"
                ),
            }
            if args.output == "json":
                print(json.dumps(payload, indent=2, sort_keys=True))
                return 0
            print(f"logs: {payload['follow_command']}")
            print(f"entries: {payload['entry_count']}")
            for record in payload["entries"]:
                timestamp = str(record.get("timestamp") or "unknown-time")
                actor = str(record.get("actor") or "session")
                message = (
                    str(record.get("message") or "").strip() or "unspecified log entry"
                )
                message_lines = message.splitlines() or ["unspecified log entry"]
                print(f"- `{timestamp}` [{actor}] {message_lines[0]}")
                for continuation in message_lines[1:]:
                    print(f"  {continuation}")
            return 0
        entries = records[-max(args.limit, 0) :] if max(args.limit, 0) > 0 else records
        payload = {
            "session_root": str(work_dir),
            "actor_count": len(actor_ids),
            "actors": actor_ids,
            "entry_count": len(records),
            "entries": entries,
            "state_paths": {
                actor: str(
                    logs_state_path(
                        session_registry._actor_session_dir(work_dir, actor)
                    )
                )
                for actor in actor_ids
            },
            "locators": {
                actor: session_charter._native_surface_locator("logs", actor_name=actor)
                for actor in actor_ids
            },
            "follow_commands": {
                actor: session_charter._native_surface_follow_command(
                    "logs", actor_name=actor
                )
                for actor in actor_ids
            },
        }
        if args.output == "json":
            print(json.dumps(payload, indent=2, sort_keys=True))
            return 0
        print(f"logs: session-wide across {len(actor_ids)} actor(s)")
        print(f"entries: {payload['entry_count']}")
        for record in payload["entries"]:
            timestamp = str(record.get("timestamp") or "unknown-time")
            actor = str(record.get("actor") or "unknown-actor")
            message = (
                str(record.get("message") or "").strip() or "unspecified log entry"
            )
            message_lines = message.splitlines() or ["unspecified log entry"]
            print(f"- `{timestamp}` [{actor}] {message_lines[0]}")
            for continuation in message_lines[1:]:
                print(f"  {continuation}")
        return 0
    work_dir = session_scope._session_dir(
        explicit_session=getattr(args, "session", None),
        explicit_actor=getattr(args, "actor", None),
    )
    writer = writer_name()
    actor_branch = work_dir.resolve().parent.name == "actors"
    target_actor = session_actor(work_dir) if actor_branch else ""
    if target_actor:
        require_writer(
            work_dir,
            target_actor,
            writer=writer,
            action="write into this actor branch",
        )
    record_actor = writer or ""
    if action == "extend":
        entries = session_charter._read_text_items_source(
            session_root=work_dir,
            inline_items=list(args.value or []),
            from_file=args.from_file,
            use_stdin=args.use_stdin,
            input_name="log entry text",
        )
        for entry in entries:
            append_log_record(
                work_dir,
                message=session_charter._normalize_entry_text(
                    entry, input_name="log entry text"
                ),
                actor=record_actor,
            )
        session_activity._record_session_activity(
            work_dir,
            plugin="logs",
            surface="logs",
            action="extend",
            actor=record_actor,
            target=logs_state_path(work_dir),
            detail=f"extended logs with {len(entries)} item(s)",
        )
        print(f"extended logs entries: {len(entries)} item(s)")
        return 0
    payload = session_charter._read_text_source(
        session_root=work_dir,
        inline=(args.value[0] if args.value else None),
        from_file=args.from_file,
        use_stdin=args.use_stdin,
        input_name="log entry text",
    )
    append_log_record(
        work_dir,
        message=session_charter._normalize_entry_text(
            payload, input_name="log entry text"
        ),
        actor=record_actor,
    )
    session_activity._record_session_activity(
        work_dir,
        plugin="logs",
        surface="logs",
        action="append",
        actor=record_actor,
        target=logs_state_path(work_dir),
        detail="appended 1 logs entry",
    )
    print("appended logs entry")
    return 0



from __future__ import annotations

from pathlib import Path
import json
from typing import Any

from gotta.content.context import current_actor
from gotta.content.env import state_dir_path
from gotta.content.file import append_jsonl_line
from gotta.content.scope import session_identity
from gotta.content.stamp import iso_utc

ACTIVITY_LOG_NAME = "activity.jsonl"


def activity_log_path(root: Path) -> Path:
    return state_dir_path(root) / ACTIVITY_LOG_NAME


def append_activity_event(root: Path, payload: dict[str, Any]) -> dict[str, Any]:
    event = dict(payload)
    event.setdefault("timestamp", iso_utc())
    event.setdefault("actor", current_actor(default_actor=session_identity(root)))
    append_jsonl_line(activity_log_path(root), event)
    return event


def activity_events(root: Path) -> list[dict[str, Any]]:
    path = activity_log_path(root)
    if not path.exists():
        return []
    records: list[dict[str, Any]] = []
    for raw_line in path.read_text(encoding="utf-8").splitlines():
        if not raw_line.strip():
            continue
        try:
            payload = json.loads(raw_line)
        except json.JSONDecodeError:
            continue
        if isinstance(payload, dict):
            records.append(payload)
    return records



"""Source metadata derivation for dispatch materialization."""

from __future__ import annotations

import json
import re
from typing import Any

from gotta.resolve.invoke import canonical_locator as resolve_canonical_locator
from gotta.source import (
    classify_visibility_metadata,
    derive_source_metadata_from_payload,
    extract_visibility_metadata_from_markdown,
    normalize_source_timestamp,
    slack_timestamp_to_iso,
)


_MARKDOWN_SOURCE_PATTERNS: tuple[tuple[re.Pattern[str], str], ...] = (
    (
        re.compile(
            r"^\s*-\s*(?:\*\*)?Created:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE
        ),
        "source_created_at",
    ),
    (
        re.compile(
            r"^\s*-\s*(?:\*\*)?Updated:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE
        ),
        "source_updated_at",
    ),
    (
        re.compile(
            r"^\s*-\s*(?:\*\*)?Modified:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE
        ),
        "source_updated_at",
    ),
    (
        re.compile(
            r"^\s*-\s*(?:\*\*)?Published:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE
        ),
        "source_published_at",
    ),
    (
        re.compile(
            r"^\s*-\s*(?:\*\*)?Authored:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE
        ),
        "source_created_at",
    ),
)


def _json_value(data: bytes) -> Any | None:
    try:
        payload = json.loads(data.decode("utf-8"))
    except (UnicodeDecodeError, json.JSONDecodeError):
        return None
    return payload


def _json_nested(payload: dict[str, Any], *path: str) -> str:
    current: Any = payload
    for key in path:
        if not isinstance(current, dict):
            return ""
        current = current.get(key)
    return str(current or "").strip()


def _extract_source_metadata_from_json(payload: Any) -> dict[str, Any]:
    metadata = derive_source_metadata_from_payload(payload)
    if not isinstance(payload, dict):
        return metadata
    candidates = (
        ("source_published_at", _json_nested(payload, "published_at")),
        ("source_published_at", _json_nested(payload, "created_at")),
        ("source_updated_at", _json_nested(payload, "updatedAt")),
        ("source_created_at", _json_nested(payload, "createdAt")),
        ("source_updated_at", _json_nested(payload, "updated")),
        ("source_created_at", _json_nested(payload, "created")),
        ("source_updated_at", _json_nested(payload, "modifiedTime")),
        ("source_created_at", _json_nested(payload, "createdTime")),
        ("source_created_at", _json_nested(payload, "commit", "author", "date")),
        ("source_created_at", _json_nested(payload, "author", "date")),
    )
    for key, value in candidates:
        parsed = normalize_source_timestamp(value) or str(value or "").strip()
        if parsed and key not in metadata:
            metadata[key] = parsed
    return metadata


def _extract_source_metadata_from_markdown(data: bytes) -> dict[str, str]:
    try:
        text = data.decode("utf-8")
    except UnicodeDecodeError:
        return {}
    metadata: dict[str, str] = {}
    for pattern, key in _MARKDOWN_SOURCE_PATTERNS:
        match = pattern.search(text)
        if not match:
            continue
        value = match.group("value").strip()
        parsed = normalize_source_timestamp(value) or value
        if parsed and key not in metadata:
            metadata[key] = parsed
    authored_match = re.search(r"\bauthored (?P<value>\d{4}-\d{2}-\d{2}T\S+Z?)\b", text)
    if authored_match and "source_created_at" not in metadata:
        metadata["source_created_at"] = authored_match.group("value")
    return metadata


def _derived_source_metadata(
    plugin: str,
    argv: list[str],
    data: bytes,
    *,
    provider: str = "",
) -> dict[str, Any]:
    metadata: dict[str, Any] = {}
    canonical = resolve_canonical_locator(plugin, argv)
    defaults: dict[str, str] = {}
    if canonical.startswith("slack:thread:"):
        thread_ts = canonical.rsplit(":", 1)[-1]
        source_time = slack_timestamp_to_iso(thread_ts)
        if source_time:
            defaults["source_created_at"] = source_time
            defaults["source_updated_at"] = source_time
    if plugin != "slack" or not argv or argv[0] != "get":
        payload = _json_value(data)
        if payload is not None:
            metadata.update(_extract_source_metadata_from_json(payload))
            metadata.update(
                classify_visibility_metadata(
                    payload,
                    provider=provider,
                    plugin=plugin,
                    subcommand=argv[0] if argv else "",
                    locator=canonical,
                )
            )
        metadata.update(
            {
                key: value
                for key, value in _extract_source_metadata_from_markdown(data).items()
                if key not in metadata
            }
        )
        metadata.update(
            {
                key: value
                for key, value in extract_visibility_metadata_from_markdown(
                    data
                ).items()
                if key not in metadata
            }
        )
        if "visibility_level" not in metadata:
            metadata.update(
                classify_visibility_metadata(
                    {},
                    provider=provider,
                    plugin=plugin,
                    subcommand=argv[0] if argv else "",
                    locator=canonical,
                )
            )
        for key, value in defaults.items():
            metadata.setdefault(key, value)
        return metadata
    payload = _json_value(data)
    if isinstance(payload, dict):
        first_ts = slack_timestamp_to_iso(str(payload.get("firstTs") or ""))
        last_ts = slack_timestamp_to_iso(str(payload.get("lastTs") or ""))
        if first_ts:
            metadata["source_created_at"] = first_ts
        if last_ts:
            metadata["source_updated_at"] = last_ts
        metadata.update(
            classify_visibility_metadata(
                payload,
                provider=provider,
                plugin=plugin,
                subcommand=argv[0] if argv else "",
                locator=canonical,
            )
        )
    metadata.update(
        {
            key: value
            for key, value in _extract_source_metadata_from_markdown(data).items()
            if key not in metadata
        }
    )
    metadata.update(
        {
            key: value
            for key, value in extract_visibility_metadata_from_markdown(data).items()
            if key not in metadata
        }
    )
    if "visibility_level" not in metadata:
        metadata.update(
            classify_visibility_metadata(
                {},
                provider=provider,
                plugin=plugin,
                subcommand=argv[0] if argv else "",
                locator=canonical,
            )
        )
    for key, value in defaults.items():
        metadata.setdefault(key, value)
    return metadata



"""Analysis surfaces for `gotta session`."""

from __future__ import annotations

import argparse
from collections import Counter
import json

from gotta.source import best_visibility_metadata
from gotta.content.model import ContentSnapshot
from gotta.content.path import content_locator
from gotta.content.store import scan_content_store
from gotta.lead.aggregate import aggregate_lead_sources
from gotta.lead.edge import build_lead_edge_records
from gotta.lead.snapshot import (
    snapshot_artifact_locator,
    snapshot_display_name,
    snapshot_locator,
)

from .core import (
    ANALYZE_ANCHOR_PREVIEW_LIMIT,
    ANALYZE_FOCUS_MATCH_PREVIEW_LIMIT,
    ANALYZE_FOCUS_NEIGHBOR_PREVIEW_LIMIT,
    SUMMARY_BUCKET_LIMIT,
    analysis_mermaid_id,
    artifact_kind,
    empty_topology_next_step,
    follow_command,
    lead_kind,
    lead_signal_labels,
    mermaid_label,
    provider_name,
    query_label,
    rendered_actor,
    render_variant,
    render_variant_label,
    resolved_visibility_metadata,
    resource_label,
    session_read_command,
    topology_next_step,
    visibility_summary,
)
from .manifest import manifest_entries
from .parse import explicit_session_ref, require_started_session, session_dirs_for_read
from .scan import scan_payload

_analysis_mermaid_id = analysis_mermaid_id
_artifact_kind = artifact_kind
_empty_topology_next_step = empty_topology_next_step
_explicit_session_ref = explicit_session_ref
_follow_command = follow_command
_lead_kind = lead_kind
_lead_signal_labels = lead_signal_labels
_manifest_entries = manifest_entries
_mermaid_label = mermaid_label
_provider_name = provider_name
_query_label = query_label
_rendered_actor = rendered_actor
_render_variant = render_variant
_render_variant_label = render_variant_label
_resolved_visibility_metadata = resolved_visibility_metadata
_resource_label = resource_label
_require_started_session = require_started_session
_scan_payload = scan_payload
_session_dirs_for_read = session_dirs_for_read
_session_read_command = session_read_command
_topology_next_step = topology_next_step
_visibility_summary = visibility_summary


def _revision_edges(snapshots: list[ContentSnapshot]) -> list[dict[str, str]]:
    tracks: dict[tuple[str, tuple[str, str]], list[dict[str, str]]] = {}
    for snapshot in snapshots:
        canonical = str(
            snapshot.metadata.get("canonical_locator", "")
            or snapshot.metadata.get("locator", "")
        ).strip()
        if not canonical:
            continue
        variant = _render_variant(snapshot)
        for event in snapshot.events:
            tracks.setdefault((canonical, variant), []).append(
                {
                    "timestamp": event.timestamp,
                    "digest": snapshot.digest,
                    "preferred_name": str(
                        snapshot.metadata.get("preferred_name", "") or event.link_name
                    ),
                    "plugin": str(snapshot.metadata.get("plugin", "") or "unknown"),
                    "actor": _rendered_actor(
                        snapshot.metadata.get("actor"),
                        session_root=snapshot.content_dir.parent.parent,
                    ),
                    "rendering": _render_variant_label(variant),
                }
            )
    edges: list[dict[str, str]] = []
    for (locator, _variant), items in sorted(tracks.items()):
        prior_item: dict[str, str] | None = None
        for item in sorted(
            items, key=lambda current: (current["timestamp"], current["digest"])
        ):
            if prior_item is None:
                prior_item = item
                continue
            if item["digest"] == prior_item["digest"]:
                prior_item = item
                continue
            edges.append(
                {
                    "locator": locator,
                    "preferredName": item["preferred_name"]
                    or prior_item["preferred_name"],
                    "from": prior_item["digest"],
                    "to": item["digest"],
                    "fromTimestamp": prior_item["timestamp"],
                    "toTimestamp": item["timestamp"],
                    "plugin": item["plugin"] or prior_item["plugin"],
                    "actor": item["actor"] or prior_item["actor"],
                    "rendering": item["rendering"] or prior_item["rendering"],
                }
            )
            prior_item = item
    return edges


def _analysis_payload(dirs, *, session_ref: str = "") -> dict[str, object]:
    snapshots = scan_content_store(dirs.content_dir)
    snapshot_by_digest = {snapshot.digest: snapshot for snapshot in snapshots}
    manifest_entries = _manifest_entries(dirs)
    source_map: dict[str, dict[str, object]] = {}
    edge_plugins: dict[tuple[str, str], list[str]] = {}
    edge_actors: dict[tuple[str, str], set[str]] = {}
    content_details: dict[str, dict[str, set[str]]] = {}

    for entry in manifest_entries:
        source = entry.get("canonical_locator") or entry.get("locator") or "unknown"
        checksum = entry.get("checksum") or ""
        if not checksum:
            continue
        source_state = source_map.setdefault(
            source,
            {
                "content": set(),
                "locators": set(),
                "plugins": set(),
                "actors": set(),
                "artifact_kinds": set(),
                "entries": 0,
                "variants": set(),
                "visibility": {},
            },
        )
        source_state["content"].add(checksum)
        locator = entry.get("locator") or source
        source_state["locators"].add(locator)
        plugin = entry.get("plugin") or "unknown"
        actor = _rendered_actor(entry.get("actor"), session_root=dirs.session_dir)
        source_state["plugins"].add(plugin)
        source_state["actors"].add(actor)
        kind = _artifact_kind(entry.get("artifact_kind"))
        if kind:
            source_state["artifact_kinds"].add(kind)
        source_state["entries"] = int(source_state["entries"]) + 1
        source_state["visibility"] = best_visibility_metadata(
            source_state.get("visibility", {}),
            _resolved_visibility_metadata(
                entry,
                provider=str(plugin),
                plugin=str(plugin),
                subcommand=str(entry.get("subcommand") or ""),
                locator=str(source),
            ),
        )
        snapshot = snapshot_by_digest.get(str(checksum))
        if snapshot is not None:
            source_state["variants"].add(_render_variant(snapshot))
        edge_plugins.setdefault((source, checksum), []).append(plugin)
        edge_actors.setdefault((source, checksum), set()).add(actor)
        detail = content_details.setdefault(
            checksum,
            {
                "providers": set(),
                "actors": set(),
                "resource_hints": set(),
            },
        )
        detail["providers"].add(
            _provider_name(source, plugins=[plugin], fallback=plugin)
        )
        detail["actors"].add(actor)
        resource_kind, resource_label = _resource_label(source)
        if resource_kind and resource_label:
            detail["resource_hints"].add(f"{resource_kind}:{resource_label}")
        else:
            detail["resource_hints"].add(source)

    name_counts = Counter(snapshot_display_name(snapshot) for snapshot in snapshots)

    content = [
        {
            "checksum": snapshot.digest,
            "preferredName": snapshot_display_name(snapshot),
            "artifactKind": _artifact_kind(snapshot.metadata.get("artifact_kind")),
            "contentLocator": content_locator(snapshot.digest),
            "artifactLocator": snapshot_artifact_locator(snapshot),
            "followCommand": _session_read_command(
                snapshot_artifact_locator(snapshot),
                session_ref=session_ref,
            ),
            "nameCollision": name_counts[snapshot_display_name(snapshot)] > 1,
            "nameCount": len(snapshot.names),
            "fetchCount": len(snapshot.events),
            "names": snapshot.names,
            "firstFetchedAt": snapshot.events[0].timestamp if snapshot.events else "",
            "lastFetchedAt": snapshot.events[-1].timestamp if snapshot.events else "",
            "providers": sorted(
                content_details.get(snapshot.digest, {}).get("providers", set())
            ),
            "actors": sorted(
                content_details.get(snapshot.digest, {}).get("actors", set())
            ),
            "resourceHints": sorted(
                content_details.get(snapshot.digest, {}).get("resource_hints", set())
            ),
            **_resolved_visibility_metadata(
                dict(snapshot.metadata),
                provider=str(snapshot.metadata.get("plugin") or ""),
                plugin=str(snapshot.metadata.get("plugin") or ""),
                subcommand=str(snapshot.metadata.get("subcommand") or ""),
                locator=str(snapshot_locator(snapshot)),
            ),
        }
        for snapshot in snapshots
    ]
    sources = [
        {
            "locator": locator,
            "contentCount": len(state["content"]),
            "entryCount": int(state["entries"]),
            "artifactKind": (
                next(iter(state["artifact_kinds"]))
                if len(state["artifact_kinds"]) == 1
                else ""
            ),
            "artifactKinds": sorted(str(value) for value in state["artifact_kinds"]),
            "plugins": sorted(str(value) for value in state["plugins"]),
            "actors": sorted(str(value) for value in state["actors"]),
            "locators": sorted(str(value) for value in state["locators"]),
            "collision": False,
            "duplicateMaterialization": len(state["content"]) > 1
            and len(state["variants"]) <= 1,
            "variant": len(state["variants"]) > 1,
            "variantCount": len(state["variants"]),
            "variants": [
                _render_variant_label(variant) for variant in sorted(state["variants"])
            ],
            **best_visibility_metadata(state.get("visibility", {})),
        }
        for locator, state in sorted(source_map.items())
    ]
    source_edges = [
        {
            "source": source,
            "checksum": checksum,
            "plugins": sorted(plugins),
            "actors": sorted(edge_actors.get((source, checksum), set())),
            "count": len(plugins),
        }
        for (source, checksum), plugins in sorted(edge_plugins.items())
    ]
    revision_edges = _revision_edges(snapshots)
    lead_edges = build_lead_edge_records(
        snapshots,
        manifest_entries,
        classify_kind=_lead_kind,
    )
    lead_sources = aggregate_lead_sources(lead_edges)
    for lead in lead_sources:
        locator = str(lead.get("locator") or "").strip()
        if locator:
            lead["followCommand"] = _follow_command(locator, session_ref=session_ref)
    collisions = [source["locator"] for source in sources if source["collision"]]
    duplicate_materializations = [
        source["locator"]
        for source in sources
        if source.get("duplicateMaterialization")
    ]
    variants = [source["locator"] for source in sources if source.get("variant")]
    name_collisions = sorted(
        name for name, count in name_counts.items() if count > 1 and name
    )
    materialized_lead_count = sum(
        1 for source in lead_sources if bool(source["materialized"])
    )
    empty = (
        not sources
        and not content
        and not source_edges
        and not revision_edges
        and not lead_edges
    )
    discovery_count = sum(
        1 for item in content if item.get("artifactKind") == "discovery"
    )
    evidence_count = sum(
        1 for item in content if item.get("artifactKind") == "evidence"
    )
    return {
        "sessionDir": str(dirs.session_dir),
        "contentDir": str(dirs.content_dir),
        "manifestPath": str(dirs.content_dir / "manifest.jsonl"),
        "manifestEntryCount": len(manifest_entries),
        "contentCount": len(content),
        "sourceCount": len(sources),
        "sourceEdgeCount": len(source_edges),
        "revisionEdgeCount": len(revision_edges),
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "collisionCount": len(collisions),
        "collisions": collisions,
        "duplicateMaterializationCount": len(duplicate_materializations),
        "duplicateMaterializations": duplicate_materializations,
        "variantCount": len(variants),
        "variants": variants,
        "nameCollisionCount": len(name_collisions),
        "nameCollisions": name_collisions,
        "leadSourceCount": len(lead_sources),
        "materializedLeadSourceCount": materialized_lead_count,
        "unmaterializedLeadSourceCount": len(lead_sources) - materialized_lead_count,
        "leadEdgeCount": len(lead_edges),
        "empty": empty,
        "nextStep": _topology_next_step(
            discovery_count=discovery_count,
            evidence_count=evidence_count,
        ),
        "sources": sources,
        "content": content,
        "sourceEdges": source_edges,
        "revisionEdges": revision_edges,
        "leadSources": lead_sources,
        "leadEdges": lead_edges,
    }


def _semantic_payload(dirs, *, session_ref: str = "") -> dict[str, object]:
    lineage = _analysis_payload(dirs, session_ref=session_ref)
    nodes: dict[str, dict[str, object]] = {}
    edges: set[tuple[str, str, str]] = set()

    def add_node(node_id: str, *, label: str, kind: str, group: str) -> None:
        nodes.setdefault(
            node_id,
            {
                "id": node_id,
                "label": label,
                "kind": kind,
                "group": group,
                "materialized": False,
                "discovered": False,
            },
        )

    def add_edge(source: str, target: str, label: str) -> None:
        edges.add((source, target, label))

    for source in lineage["sources"]:
        locator = str(source["locator"])
        provider = _provider_name(
            locator,
            plugins=[str(value) for value in source.get("plugins") or [] if str(value)],
        )
        provider_id = f"provider:{provider}"
        source_id = f"source:{locator}"
        add_node(provider_id, label=provider, kind="provider", group=provider)
        add_node(source_id, label=locator, kind="source", group=provider)
        nodes[source_id]["materialized"] = True
        nodes[source_id]["artifactKind"] = str(source.get("artifactKind") or "")
        nodes[source_id]["artifactKinds"] = list(source.get("artifactKinds") or [])
        add_edge(provider_id, source_id, "source")

        query_label = _query_label(locator)
        if query_label:
            query_id = f"query:{provider}:{query_label}"
            add_node(query_id, label=query_label, kind="query", group=provider)
            add_edge(provider_id, query_id, "query")
            add_edge(query_id, source_id, "drives")

        resource_kind, resource_label = _resource_label(locator)
        if resource_kind and resource_label:
            resource_id = f"resource:{provider}:{resource_label}"
            add_node(
                resource_id,
                label=resource_label,
                kind=resource_kind,
                group=provider,
            )
            add_edge(provider_id, resource_id, "resource")
            add_edge(resource_id, source_id, "resolved_by")

    for content in lineage["content"]:
        checksum = str(content["checksum"])
        content_id = f"content:{checksum}"
        add_node(
            content_id,
            label=str(content["preferredName"]),
            kind="content",
            group="content",
        )
        nodes[content_id]["materialized"] = True
        nodes[content_id]["artifactKind"] = str(content.get("artifactKind") or "")

    for edge in lineage["sourceEdges"]:
        source_id = f"source:{edge['source']}"
        content_id = f"content:{edge['checksum']}"
        add_edge(
            source_id, content_id, ",".join(str(value) for value in edge["plugins"])
        )

    for edge in lineage["revisionEdges"]:
        from_id = f"content:{edge['from']}"
        to_id = f"content:{edge['to']}"
        add_edge(from_id, to_id, f"revision:{edge['locator']}")
    for lead_source in lineage.get("leadSources") or []:
        locator = str(lead_source["locator"])
        provider = str(lead_source["provider"] or _provider_name(locator))
        provider_id = f"provider:{provider}"
        source_id = f"source:{locator}"
        add_node(provider_id, label=provider, kind="provider", group=provider)
        add_node(source_id, label=locator, kind="source", group=provider)
        nodes[source_id]["materialized"] = bool(
            nodes[source_id].get("materialized") or lead_source.get("materialized")
        )
        nodes[source_id]["discovered"] = True
        if not nodes[source_id].get("artifactKinds"):
            nodes[source_id]["artifactKinds"] = []
        kinds = {
            str(value)
            for value in nodes[source_id].get("artifactKinds") or []
            if str(value)
        }
        lead_kind = str(lead_source.get("artifactKind") or "")
        if lead_kind:
            kinds.add(lead_kind)
        nodes[source_id]["artifactKinds"] = sorted(kinds)
        nodes[source_id]["artifactKind"] = (
            nodes[source_id]["artifactKinds"][0]
            if len(nodes[source_id]["artifactKinds"]) == 1
            else ""
        )
        add_edge(provider_id, source_id, "source")
        resource_kind, resource_label = _resource_label(locator)
        if resource_kind and resource_label:
            resource_id = f"resource:{provider}:{resource_label}"
            add_node(
                resource_id,
                label=resource_label,
                kind=resource_kind,
                group=provider,
            )
            add_edge(provider_id, resource_id, "resource")
            add_edge(resource_id, source_id, "resolved_by")
    for edge in lineage.get("leadEdges") or []:
        from_id = f"content:{edge['sourceChecksum']}"
        to_id = f"source:{edge['targetLocator']}"
        relation = str(edge.get("relation") or "links_to")
        count = int(edge.get("occurrenceCount") or 0)
        add_edge(from_id, to_id, relation if count <= 1 else f"{relation} x{count}")

    empty = not nodes and not edges
    discovery_count = int(lineage.get("discoveryArtifactCount") or 0)
    evidence_count = int(lineage.get("evidenceArtifactCount") or 0)
    return {
        "sessionDir": lineage["sessionDir"],
        "contentDir": lineage["contentDir"],
        "manifestPath": lineage["manifestPath"],
        "nodeCount": len(nodes),
        "edgeCount": len(edges),
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "empty": empty,
        "nextStep": _topology_next_step(
            discovery_count=discovery_count,
            evidence_count=evidence_count,
        ),
        "nodes": sorted(nodes.values(), key=lambda item: (item["kind"], item["label"])),
        "edges": [
            {"source": source, "target": target, "label": label}
            for source, target, label in sorted(edges)
        ],
    }


def _render_semantic_mermaid(payload: dict[str, object]) -> str:
    lines = [
        "---",
        "title: gotta semantic graph",
        "---",
        "flowchart LR",
    ]
    if payload.get("empty"):
        lines.append(
            f'  empty["{_mermaid_label(str(payload.get("nextStep") or _empty_topology_next_step()))}"]'
        )
        lines.extend(
            [
                "  class empty emptyState",
                "  classDef emptyState fill:#f7fafc,stroke:#4a5568,color:#1a202c;",
                "",
            ]
        )
        return "\n".join(lines)
    if payload.get("nextStep"):
        lines.append(f'  note["{_mermaid_label(str(payload["nextStep"]))}"]')
        lines.append("  class note emptyState")
    for node in payload["nodes"]:
        node_id = _analysis_mermaid_id("sem", str(node["id"]))
        label = _mermaid_label(str(node["label"]))
        kind = str(node["kind"]).replace("-", "_")
        if (
            str(node.get("kind")) == "source"
            and not bool(node.get("materialized"))
            and bool(node.get("discovered"))
        ):
            kind = "source_gap"
        lines.append(f'  {node_id}["{label}"]')
        lines.append(f"  class {node_id} {kind}")
    for edge in payload["edges"]:
        source_id = _analysis_mermaid_id("sem", str(edge["source"]))
        target_id = _analysis_mermaid_id("sem", str(edge["target"]))
        label = _mermaid_label(str(edge["label"]))
        lines.append(f"  {source_id} -->|{label}| {target_id}")
    lines.extend(
        [
            "  classDef provider fill:#fff7e6,stroke:#b7791f,color:#5a3b09;",
            "  classDef query fill:#f5f3ff,stroke:#6b46c1,color:#34206b;",
            "  classDef source fill:#eef8ee,stroke:#2d7a2d,color:#173d17;",
            "  classDef source_gap fill:#fffaf0,stroke:#b7791f,color:#5a3b09;",
            "  classDef content fill:#eef4ff,stroke:#2a62c7,color:#173058;",
            "  classDef jira_issue fill:#fff5f5,stroke:#c53030,color:#63171b;",
            "  classDef confluence_page fill:#fffbea,stroke:#b7791f,color:#5a3b09;",
            "  classDef google_doc fill:#effcf6,stroke:#2f855a,color:#1b4332;",
            "  classDef google_drive_file fill:#effcf6,stroke:#2f855a,color:#1b4332;",
            "  classDef slack_thread fill:#ebf8ff,stroke:#2b6cb0,color:#1a365d;",
            "  classDef slack_channel fill:#ebf8ff,stroke:#2b6cb0,color:#1a365d;",
            "  classDef github_repo fill:#f7fafc,stroke:#4a5568,color:#1a202c;",
            "  classDef emptyState fill:#f7fafc,stroke:#4a5568,color:#1a202c;",
            "",
        ]
    )
    return "\n".join(lines)


def _render_analysis_mermaid(payload: dict[str, object]) -> str:
    lines = [
        "---",
        "title: gotta session analysis",
        "---",
        "flowchart LR",
    ]
    if payload.get("empty"):
        lines.append(
            f'  empty["{_mermaid_label(str(payload.get("nextStep") or _empty_topology_next_step()))}"]'
        )
        lines.extend(
            [
                "  class empty emptyState",
                "  classDef emptyState fill:#f7fafc,stroke:#4a5568,color:#1a202c;",
                "",
            ]
        )
        return "\n".join(lines)
    if payload.get("nextStep"):
        lines.append(f'  note["{_mermaid_label(str(payload["nextStep"]))}"]')
        lines.append("  class note emptyState")
    for source in payload["sources"]:
        locator = str(source["locator"])
        node_id = _analysis_mermaid_id("src", locator)
        actors = ", ".join(str(value) for value in source.get("actors") or [])
        label_parts = [locator]
        if actors:
            label_parts.append(f"actor: {actors}")
        if source.get("variant"):
            label_parts.append(f"renderings: {int(source.get('variantCount') or 0)}")
        elif source.get("duplicateMaterialization"):
            label_parts.append(
                f"materializations: {int(source.get('contentCount') or 0)}"
            )
        label = _mermaid_label("\n".join(label_parts))
        lines.append(f'  {node_id}["{label}"]')
        if source["collision"]:
            lines.append(f"  class {node_id} collision")
        elif source.get("variant"):
            lines.append(f"  class {node_id} variant")
        elif source.get("duplicateMaterialization"):
            lines.append(f"  class {node_id} duplicate")
        else:
            lines.append(f"  class {node_id} source")
    for content in payload["content"]:
        checksum = str(content["checksum"])
        providers = ", ".join(str(value) for value in content.get("providers") or [])
        actors = ", ".join(str(value) for value in content.get("actors") or [])
        label_parts = [str(content["preferredName"])]
        resource_hints = [
            str(value) for value in content.get("resourceHints") or [] if str(value)
        ]
        if bool(content.get("nameCollision")) and resource_hints:
            label_parts.append(resource_hints[0])
        if providers:
            label_parts.append(providers)
        if actors:
            label_parts.append(f"actor: {actors}")
        label_parts.append(checksum[:12])
        label = _mermaid_label("\n".join(label_parts))
        node_id = _analysis_mermaid_id("art", checksum)
        lines.append(f'  {node_id}["{label}"]')
        lines.append(f"  class {node_id} content")
    for edge in payload["sourceEdges"]:
        source_id = _analysis_mermaid_id("src", str(edge["source"]))
        content_id = _analysis_mermaid_id("art", str(edge["checksum"]))
        label_parts = [", ".join(edge["plugins"])]
        actors = ", ".join(str(value) for value in edge.get("actors") or [])
        if actors:
            label_parts.append(f"actor: {actors}")
        label = _mermaid_label("\n".join(part for part in label_parts if part))
        lines.append(f"  {source_id} -->|{label}| {content_id}")
    for edge in payload["revisionEdges"]:
        from_id = _analysis_mermaid_id("art", str(edge["from"]))
        to_id = _analysis_mermaid_id("art", str(edge["to"]))
        label = _mermaid_label(
            f"revision:{str(edge['locator'])}\n{str(edge.get('rendering') or '')}".rstrip()
        )
        lines.append(f"  {from_id} -->|{label}| {to_id}")
    seen_source_nodes = {str(source["locator"]) for source in payload["sources"]}
    for lead_source in payload.get("leadSources") or []:
        locator = str(lead_source["locator"])
        if locator in seen_source_nodes:
            continue
        node_id = _analysis_mermaid_id("src", locator)
        label_parts = [locator, f"lead: {str(lead_source['provider'])}"]
        if not bool(lead_source.get("materialized")):
            label_parts.append("not yet materialized")
        label = _mermaid_label("\n".join(label_parts))
        lines.append(f'  {node_id}["{label}"]')
        lines.append(f"  class {node_id} leadgap")
    for edge in payload.get("leadEdges") or []:
        content_id = _analysis_mermaid_id("art", str(edge["sourceChecksum"]))
        source_id = _analysis_mermaid_id("src", str(edge["targetLocator"]))
        relation = str(edge.get("relation") or "links_to")
        count = int(edge.get("occurrenceCount") or 0)
        label = relation if count <= 1 else f"{relation} x{count}"
        lines.append(f"  {content_id} -.->|{_mermaid_label(label)}| {source_id}")
    lines.extend(
        [
            "  classDef source fill:#eef8ee,stroke:#2d7a2d,color:#173d17;",
            "  classDef content fill:#eef4ff,stroke:#2a62c7,color:#173058;",
            "  classDef duplicate fill:#edf2f7,stroke:#4a5568,color:#1a202c;",
            "  classDef variant fill:#fff7e6,stroke:#b7791f,color:#5a3b09;",
            "  classDef collision fill:#fff1f1,stroke:#c73434,color:#6b1111;",
            "  classDef leadgap fill:#fffaf0,stroke:#b7791f,color:#5a3b09;",
            "  classDef emptyState fill:#f7fafc,stroke:#4a5568,color:#1a202c;",
            "",
        ]
    )
    return "\n".join(lines)


def _semantic_node_follow_command(
    node: dict[str, object],
    *,
    lineage: dict[str, object],
) -> str:
    kind = str(node.get("kind") or "")
    node_id = str(node.get("id") or "")
    label = str(node.get("label") or "").strip()
    if kind == "source" and label:
        return _follow_command(label)
    if kind == "content" and node_id.startswith("content:"):
        checksum = node_id.split(":", 1)[1]
        for content_item in lineage.get("content") or []:
            if str(content_item.get("checksum") or "") == checksum:
                return str(content_item.get("followCommand") or "").strip()
    return ""


def _focus_match_threshold(best_score: int) -> int:
    if best_score <= 0:
        return 0
    if best_score >= 4:
        return 2
    return best_score


def _ordered_focus_scan_entries(
    scan_payload: dict[str, object] | None,
    *,
    limit: int,
) -> list[dict[str, object]]:
    if not isinstance(scan_payload, dict):
        return []
    entries = [
        dict(entry)
        for entry in scan_payload.get("entries") or []
        if isinstance(entry, dict)
    ]
    ordered = sorted(
        entries,
        key=lambda entry: str(
            entry.get("lastFetchedAt") or entry.get("fetched_at") or ""
        ),
        reverse=True,
    )
    ordered = sorted(
        ordered,
        key=lambda entry: int(entry.get("hitCount") or 0),
        reverse=True,
    )
    ordered = sorted(
        ordered,
        key=lambda entry: (
            str(entry.get("artifactKind") or entry.get("artifact_kind") or "")
            != "evidence"
        ),
    )
    return ordered[: max(limit, 0)]


def _lineage_source_candidate(item: dict[str, object]) -> dict[str, object]:
    return {
        "kind": "source",
        "label": str(item.get("locator") or ""),
        "locator": str(item.get("locator") or ""),
        "artifactKind": str(item.get("artifactKind") or ""),
        "materialized": True,
        "followCommand": str(item.get("followCommand") or ""),
    }


def _lineage_content_candidate(item: dict[str, object]) -> dict[str, object]:
    return {
        "kind": "content",
        "label": str(item.get("preferredName") or ""),
        "checksum": str(item.get("checksum") or ""),
        "artifactLocator": str(item.get("artifactLocator") or ""),
        "contentLocator": str(item.get("contentLocator") or ""),
        "artifactKind": str(item.get("artifactKind") or ""),
        "materialized": True,
        "followCommand": str(item.get("followCommand") or ""),
    }


def _lineage_lead_candidate(item: dict[str, object]) -> dict[str, object]:
    return {
        "kind": "lead",
        "label": str(item.get("locator") or ""),
        "locator": str(item.get("locator") or ""),
        "artifactKind": str(item.get("artifactKind") or ""),
        "materialized": bool(item.get("materialized")),
        "followCommand": str(item.get("followCommand") or ""),
    }


def _analysis_focus_score(
    node: dict[str, object], query: str
) -> tuple[int, int, int, str]:
    query_lower = query.lower()
    label = str(node.get("label") or "")
    node_id = str(node.get("id") or "")
    label_lower = label.lower()
    node_id_lower = node_id.lower()
    score = 0
    if label_lower == query_lower or node_id_lower == query_lower:
        score = 5
    elif label_lower.startswith(query_lower) or node_id_lower.startswith(query_lower):
        score = 4
    elif f":{query_lower}" in node_id_lower:
        score = 3
    elif query_lower in label_lower or query_lower in node_id_lower:
        score = 2
    materialized = 1 if bool(node.get("materialized")) else 0
    discovered = 1 if bool(node.get("discovered")) else 0
    return (score, materialized, discovered, label_lower)


def _neighbor_sort_key(
    node: dict[str, object],
    *,
    relation_labels: list[str],
) -> tuple[int, int, int, str, str]:
    interesting_relations = sum(
        1
        for label in relation_labels
        if label not in {"source", "resource", "resolved_by", "query", "drives"}
    )
    return (
        interesting_relations,
        1 if bool(node.get("materialized")) else 0,
        1 if bool(node.get("discovered")) else 0,
        str(node.get("kind") or ""),
        str(node.get("label") or "").lower(),
    )


def _semantic_focus_payload(
    lineage: dict[str, object],
    semantic: dict[str, object],
    *,
    focus: str,
    limit: int,
    scan_payload: dict[str, object] | None = None,
) -> dict[str, object]:
    query = focus.strip()
    if not query:
        return {
            "sessionDir": semantic["sessionDir"],
            "contentDir": semantic["contentDir"],
            "focus": "",
            "matched": False,
            "empty": True,
            "nextStep": "Provide a focus keyword, locator, artifact name, or checksum prefix.",
            "nodeCount": 0,
            "edgeCount": 0,
            "nodes": [],
            "edges": [],
            "neighbors": [],
            "anchors": [],
            "matchedCount": 0,
        }
    nodes = [dict(node) for node in semantic.get("nodes") or []]
    node_index = {str(node["id"]): node for node in nodes}
    matches = sorted(
        (
            node
            for node in nodes
            if str(node.get("kind") or "") in {"source", "content"}
            and _analysis_focus_score(node, query)[0] > 0
        ),
        key=lambda node: _analysis_focus_score(node, query),
        reverse=True,
    )
    scan_entries = _ordered_focus_scan_entries(
        scan_payload,
        limit=max(limit * 2, 8),
    )
    seed_cap = max(4, min(max(limit, 1), 12))
    if not matches and not scan_entries:
        return {
            "sessionDir": semantic["sessionDir"],
            "contentDir": semantic["contentDir"],
            "focus": query,
            "matched": False,
            "empty": True,
            "nextStep": (
                f"No analyzed node or projected artifact matched `{query}`. Try a canonical locator, "
                "artifact name, checksum prefix, or a tighter keyword from session scan, leads, or manifest."
            ),
            "nodeCount": 0,
            "edgeCount": 0,
            "nodes": [],
            "edges": [],
            "neighbors": [],
            "anchors": [],
            "matchedCount": 0,
        }

    best_score = _analysis_focus_score(matches[0], query)[0] if matches else 0
    threshold = _focus_match_threshold(best_score)
    seed_ids: list[str] = []

    def add_seed(node_id: str) -> None:
        if node_id and node_id in node_index and node_id not in seed_ids:
            seed_ids.append(node_id)

    for node in matches:
        if _analysis_focus_score(node, query)[0] < threshold:
            break
        add_seed(str(node.get("id") or ""))
        if len(seed_ids) >= seed_cap:
            break
    for entry in scan_entries:
        checksum = str(entry.get("checksum") or "").strip()
        locator = str(
            entry.get("canonical_locator") or entry.get("locator") or ""
        ).strip()
        if checksum:
            add_seed(f"content:{checksum}")
        if locator:
            add_seed(f"source:{locator}")
        if len(seed_ids) >= seed_cap:
            break

    if not seed_ids:
        return {
            "sessionDir": semantic["sessionDir"],
            "contentDir": semantic["contentDir"],
            "focus": query,
            "matched": False,
            "empty": True,
            "nextStep": (
                f"No analyzed node or projected artifact matched `{query}`. Try a canonical locator, "
                "artifact name, checksum prefix, or a tighter keyword from session scan, leads, or manifest."
            ),
            "nodeCount": 0,
            "edgeCount": 0,
            "nodes": [],
            "edges": [],
            "neighbors": [],
            "anchors": [],
            "matchedCount": 0,
        }

    root = dict(node_index[seed_ids[0]])
    root["followCommand"] = _semantic_node_follow_command(root, lineage=lineage)
    root_id = str(root["id"])
    seed_records = []
    for node_id in seed_ids:
        node = dict(node_index[node_id])
        node["followCommand"] = _semantic_node_follow_command(node, lineage=lineage)
        seed_records.append(node)
    seed_id_set = set(seed_ids)
    structural_labels = {"source", "resource", "resolved_by", "query", "drives"}
    incident_edges = [
        dict(edge)
        for edge in semantic.get("edges") or []
        if (
            str(edge.get("source") or "") in seed_id_set
            or str(edge.get("target") or "") in seed_id_set
        )
    ]
    semantic_incident_edges = [
        edge
        for edge in incident_edges
        if str(edge.get("label") or "") not in structural_labels
    ]
    selected_edges = semantic_incident_edges or incident_edges
    selected_neighbor_ids: list[str] = []
    relation_labels_by_neighbor: dict[str, list[str]] = {}
    for edge in selected_edges:
        source = str(edge.get("source") or "")
        target = str(edge.get("target") or "")
        if source in seed_id_set and target in seed_id_set:
            continue
        neighbor_id = target if source in seed_id_set else source
        if not neighbor_id:
            continue
        if neighbor_id not in relation_labels_by_neighbor:
            relation_labels_by_neighbor[neighbor_id] = []
        relation_labels_by_neighbor[neighbor_id].append(str(edge.get("label") or ""))
        if neighbor_id not in selected_neighbor_ids:
            selected_neighbor_ids.append(neighbor_id)
    if len(selected_neighbor_ids) < max(2, limit // 2):
        for edge in incident_edges:
            source = str(edge.get("source") or "")
            target = str(edge.get("target") or "")
            if source in seed_id_set and target in seed_id_set:
                continue
            neighbor_id = target if source in seed_id_set else source
            if not neighbor_id:
                continue
            if neighbor_id not in relation_labels_by_neighbor:
                relation_labels_by_neighbor[neighbor_id] = []
            relation_labels_by_neighbor[neighbor_id].append(
                str(edge.get("label") or "")
            )
            if neighbor_id not in selected_neighbor_ids:
                selected_neighbor_ids.append(neighbor_id)
    selected_neighbor_ids = [
        neighbor_id
        for neighbor_id in sorted(
            selected_neighbor_ids,
            key=lambda neighbor_id: _neighbor_sort_key(
                node_index.get(neighbor_id, {}),
                relation_labels=relation_labels_by_neighbor.get(neighbor_id, []),
            ),
            reverse=True,
        )[: max(limit, 0)]
        if neighbor_id in node_index
    ]
    selected_node_ids = {root_id, *seed_ids, *selected_neighbor_ids}
    focused_edges = [
        edge
        for edge in incident_edges
        if (
            str(edge.get("source") or "") in selected_node_ids
            and str(edge.get("target") or "") in selected_node_ids
        )
    ]
    neighbor_records = []
    for neighbor_id in selected_neighbor_ids:
        node = dict(node_index[neighbor_id])
        node["followCommand"] = _semantic_node_follow_command(node, lineage=lineage)
        relation_labels = relation_labels_by_neighbor.get(neighbor_id, [])
        neighbor_records.append(
            {
                **node,
                "relations": relation_labels,
            }
        )
    focused_nodes = [*seed_records, *neighbor_records]
    suppressed_count = max(len(incident_edges) - len(focused_edges), 0)
    return {
        "sessionDir": semantic["sessionDir"],
        "contentDir": semantic["contentDir"],
        "focus": query,
        "matched": True,
        "empty": False,
        "nextStep": "",
        "nodeCount": len(focused_nodes),
        "edgeCount": len(focused_edges),
        "root": root,
        "anchors": seed_records[1:],
        "matchedCount": len(seed_records),
        "neighbors": neighbor_records,
        "nodes": focused_nodes,
        "edges": focused_edges,
        "suppressedStructuralEdgeCount": suppressed_count,
    }


def _lineage_focus_score(item: dict[str, object], query: str) -> tuple[int, int, str]:
    query_lower = query.lower()
    candidates = [
        str(item.get("label") or ""),
        str(item.get("locator") or ""),
        str(item.get("artifactLocator") or ""),
        str(item.get("contentLocator") or ""),
        str(item.get("checksum") or ""),
    ]
    lowered = [candidate.lower() for candidate in candidates if candidate]
    score = 0
    if any(candidate == query_lower for candidate in lowered):
        score = 5
    elif any(candidate.startswith(query_lower) for candidate in lowered):
        score = 4
    elif any(f":{query_lower}" in candidate for candidate in lowered):
        score = 3
    elif any(query_lower in candidate for candidate in lowered):
        score = 2
    materialized = 1 if bool(item.get("materialized")) else 0
    label = str(item.get("label") or item.get("locator") or "")
    return (score, materialized, label.lower())


def _lineage_focus_payload(
    payload: dict[str, object],
    *,
    focus: str,
    limit: int,
    scan_payload: dict[str, object] | None = None,
) -> dict[str, object]:
    query = focus.strip()
    if not query:
        return {
            "sessionDir": payload["sessionDir"],
            "contentDir": payload["contentDir"],
            "focus": "",
            "matched": False,
            "empty": True,
            "nextStep": "Provide a focus keyword, locator, artifact name, or checksum prefix.",
            "root": {},
            "neighbors": [],
            "sources": [],
            "content": [],
            "sourceEdges": [],
            "revisionEdges": [],
            "leadSources": [],
            "leadEdges": [],
            "discoveryArtifactCount": 0,
            "evidenceArtifactCount": 0,
            "anchors": [],
            "matchedCount": 0,
        }
    sources = [dict(item) for item in payload.get("sources") or []]
    content_items = [dict(item) for item in payload.get("content") or []]
    lead_sources = [dict(item) for item in payload.get("leadSources") or []]
    source_index = {str(item.get("locator") or ""): item for item in sources}
    content_index = {str(item.get("checksum") or ""): item for item in content_items}
    lead_index = {str(item.get("locator") or ""): item for item in lead_sources}
    candidates = [
        *(_lineage_source_candidate(item) for item in sources),
        *(_lineage_content_candidate(item) for item in content_items),
        *(
            _lineage_lead_candidate(item)
            for item in lead_sources
            if str(item.get("locator") or "") not in source_index
        ),
    ]
    matches = sorted(
        (
            candidate
            for candidate in candidates
            if _lineage_focus_score(candidate, query)[0] > 0
        ),
        key=lambda candidate: _lineage_focus_score(candidate, query),
        reverse=True,
    )
    scan_entries = _ordered_focus_scan_entries(
        scan_payload,
        limit=max(limit * 2, 8),
    )
    seed_cap = max(4, min(max(limit, 1), 12))
    if not matches and not scan_entries:
        return {
            "sessionDir": payload["sessionDir"],
            "contentDir": payload["contentDir"],
            "focus": query,
            "matched": False,
            "empty": True,
            "nextStep": (
                f"No analyzed lineage anchor or projected artifact matched `{query}`. Try a canonical locator, "
                "artifact name, checksum prefix, or a tighter target from session scan, leads, or manifest."
            ),
            "root": {},
            "neighbors": [],
            "sources": [],
            "content": [],
            "sourceEdges": [],
            "revisionEdges": [],
            "leadSources": [],
            "leadEdges": [],
            "discoveryArtifactCount": 0,
            "evidenceArtifactCount": 0,
            "anchors": [],
            "matchedCount": 0,
        }
    best_score = _lineage_focus_score(matches[0], query)[0] if matches else 0
    threshold = _focus_match_threshold(best_score)
    seeds: list[dict[str, object]] = []
    seen_seed_keys: set[tuple[str, str]] = set()

    def add_seed(candidate: dict[str, object]) -> None:
        kind = str(candidate.get("kind") or "")
        if kind == "source":
            key = ("source", str(candidate.get("locator") or ""))
        elif kind == "content":
            key = ("content", str(candidate.get("checksum") or ""))
        else:
            key = ("lead", str(candidate.get("locator") or ""))
        if not key[1] or key in seen_seed_keys:
            return
        seen_seed_keys.add(key)
        seeds.append(candidate)

    for candidate in matches:
        if _lineage_focus_score(candidate, query)[0] < threshold:
            break
        add_seed(dict(candidate))
        if len(seeds) >= seed_cap:
            break
    for entry in scan_entries:
        checksum = str(entry.get("checksum") or "").strip()
        locator = str(
            entry.get("canonical_locator") or entry.get("locator") or ""
        ).strip()
        if checksum and checksum in content_index:
            add_seed(_lineage_content_candidate(content_index[checksum]))
        if locator and locator in source_index:
            add_seed(_lineage_source_candidate(source_index[locator]))
        if len(seeds) >= seed_cap:
            break

    if not seeds:
        return {
            "sessionDir": payload["sessionDir"],
            "contentDir": payload["contentDir"],
            "focus": query,
            "matched": False,
            "empty": True,
            "nextStep": (
                f"No analyzed lineage anchor or projected artifact matched `{query}`. Try a canonical locator, "
                "artifact name, checksum prefix, or a tighter target from session scan, leads, or manifest."
            ),
            "root": {},
            "neighbors": [],
            "sources": [],
            "content": [],
            "sourceEdges": [],
            "revisionEdges": [],
            "leadSources": [],
            "leadEdges": [],
            "discoveryArtifactCount": 0,
            "evidenceArtifactCount": 0,
            "anchors": [],
            "matchedCount": 0,
        }

    root = dict(seeds[0])
    selected_sources: set[str] = set()
    selected_content: set[str] = set()
    selected_leads: set[str] = set()
    matched_sources = {
        str(candidate.get("locator") or "")
        for candidate in matches
        if str(candidate.get("kind") or "") == "source"
    }
    matched_leads = {
        str(candidate.get("locator") or "")
        for candidate in matches
        if str(candidate.get("kind") or "") == "lead"
    }
    for candidate in seeds:
        if candidate["kind"] == "source":
            selected_sources.add(str(candidate.get("locator") or ""))
        elif candidate["kind"] == "content":
            selected_content.add(str(candidate.get("checksum") or ""))
        else:
            selected_leads.add(str(candidate.get("locator") or ""))

    def expand_source_and_revision_edges() -> None:
        for edge in payload.get("sourceEdges") or []:
            source = str(edge.get("source") or "")
            checksum = str(edge.get("checksum") or "")
            if source in selected_sources:
                selected_content.add(checksum)
            if checksum in selected_content:
                selected_sources.add(source)
        for edge in payload.get("revisionEdges") or []:
            from_checksum = str(edge.get("from") or "")
            to_checksum = str(edge.get("to") or "")
            if from_checksum in selected_content:
                selected_content.add(to_checksum)
            if to_checksum in selected_content:
                selected_content.add(from_checksum)

    expand_source_and_revision_edges()
    for edge in payload.get("leadEdges") or []:
        source_checksum = str(edge.get("sourceChecksum") or "")
        target_locator = str(edge.get("targetLocator") or "")
        target_is_source = target_locator in source_index
        target_matches_focus = (
            target_locator in matched_sources or target_locator in matched_leads
        )
        if source_checksum in selected_content and (
            target_locator in selected_sources
            or target_locator in selected_leads
            or target_matches_focus
        ):
            if target_is_source:
                selected_sources.add(target_locator)
            else:
                selected_leads.add(target_locator)
        if target_locator in selected_sources or target_locator in selected_leads:
            selected_content.add(source_checksum)
    expand_source_and_revision_edges()

    neighbor_candidates: list[dict[str, object]] = []
    for locator in sorted(selected_sources):
        if ("source", locator) in seen_seed_keys:
            continue
        source_item = source_index.get(locator)
        if source_item is None:
            continue
        neighbor_candidates.append(
            {
                "kind": "source",
                "label": locator,
                "relation": "materialized source",
                "followCommand": str(source_item.get("followCommand") or ""),
                "artifactKind": str(source_item.get("artifactKind") or ""),
                "materialized": True,
            }
        )
    for checksum in sorted(selected_content):
        if ("content", checksum) in seen_seed_keys:
            continue
        content_item = content_index.get(checksum)
        if content_item is None:
            continue
        neighbor_candidates.append(
            {
                "kind": "content",
                "label": str(content_item.get("preferredName") or checksum),
                "relation": "stored artifact",
                "followCommand": str(content_item.get("followCommand") or ""),
                "artifactKind": str(content_item.get("artifactKind") or ""),
                "materialized": True,
            }
        )
    for locator in sorted(selected_leads):
        if ("lead", locator) in seen_seed_keys:
            continue
        lead_item = lead_index.get(locator)
        if lead_item is None:
            continue
        neighbor_candidates.append(
            {
                "kind": "lead",
                "label": locator,
                "relation": "followable lead",
                "followCommand": str(lead_item.get("followCommand") or ""),
                "artifactKind": str(lead_item.get("artifactKind") or ""),
                "materialized": bool(lead_item.get("materialized")),
            }
        )
    ordered_neighbors = sorted(
        neighbor_candidates,
        key=lambda item: (
            1 if bool(item.get("materialized")) else 0,
            1 if str(item.get("kind") or "") == "content" else 0,
            str(item.get("label") or "").lower(),
        ),
        reverse=True,
    )[: max(limit, 0)]

    neighbor_source_labels = {
        str(item.get("label") or "")
        for item in ordered_neighbors
        if str(item.get("kind") or "") == "source"
    }
    neighbor_content_labels = {
        str(item.get("label") or "")
        for item in ordered_neighbors
        if str(item.get("kind") or "") == "content"
    }
    neighbor_content_checksums = {
        checksum
        for checksum, item in content_index.items()
        if str(item.get("preferredName") or checksum) in neighbor_content_labels
    }
    neighbor_lead_labels = {
        str(item.get("label") or "")
        for item in ordered_neighbors
        if str(item.get("kind") or "") == "lead"
    }
    seed_source_labels = {value for kind, value in seen_seed_keys if kind == "source"}
    seed_content_checksums = {
        value for kind, value in seen_seed_keys if kind == "content"
    }
    seed_lead_labels = {value for kind, value in seen_seed_keys if kind == "lead"}
    selected_sources = {
        locator
        for locator in selected_sources
        if locator in seed_source_labels or locator in neighbor_source_labels
    }
    selected_content = {
        checksum
        for checksum in selected_content
        if checksum in seed_content_checksums or checksum in neighbor_content_checksums
    }
    selected_leads = {
        locator
        for locator in selected_leads
        if locator in seed_lead_labels or locator in neighbor_lead_labels
    }

    selected_source_items = [
        item for item in sources if str(item.get("locator") or "") in selected_sources
    ]
    selected_content_items = [
        item
        for item in content_items
        if str(item.get("checksum") or "") in selected_content
    ]
    selected_lead_items = [
        item
        for item in lead_sources
        if str(item.get("locator") or "") in selected_leads
        and str(item.get("locator") or "") not in selected_sources
    ]
    selected_source_edges = [
        edge
        for edge in payload.get("sourceEdges") or []
        if str(edge.get("source") or "") in selected_sources
        and str(edge.get("checksum") or "") in selected_content
    ]
    selected_revision_edges = [
        edge
        for edge in payload.get("revisionEdges") or []
        if str(edge.get("from") or "") in selected_content
        and str(edge.get("to") or "") in selected_content
    ]
    selected_lead_edges = [
        edge
        for edge in payload.get("leadEdges") or []
        if str(edge.get("sourceChecksum") or "") in selected_content
        and str(edge.get("targetLocator") or "")
        in selected_leads.union(selected_sources)
    ]
    discovery_count = sum(
        1
        for item in selected_content_items
        if str(item.get("artifactKind") or "") == "discovery"
    )
    evidence_count = sum(
        1
        for item in selected_content_items
        if str(item.get("artifactKind") or "") == "evidence"
    )
    return {
        "sessionDir": payload["sessionDir"],
        "contentDir": payload["contentDir"],
        "manifestPath": payload["manifestPath"],
        "focus": query,
        "matched": True,
        "empty": False,
        "nextStep": "",
        "root": root,
        "anchors": seeds[1:],
        "matchedCount": len(seeds),
        "neighbors": ordered_neighbors,
        "sources": selected_source_items,
        "content": selected_content_items,
        "sourceEdges": selected_source_edges,
        "revisionEdges": selected_revision_edges,
        "leadSources": selected_lead_items,
        "leadEdges": selected_lead_edges,
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "contentCount": len(selected_content_items),
        "sourceCount": len(selected_source_items),
        "sourceEdgeCount": len(selected_source_edges),
        "revisionEdgeCount": len(selected_revision_edges),
        "leadSourceCount": len(selected_lead_items),
        "leadEdgeCount": len(selected_lead_edges),
        "collisionCount": 0,
        "collisions": [],
        "duplicateMaterializationCount": 0,
        "duplicateMaterializations": [],
        "variantCount": 0,
        "variants": [],
    }


def _analysis_overview_payload(
    lineage: dict[str, object],
    semantic: dict[str, object],
    *,
    limit: int,
) -> dict[str, object]:
    node_groups = Counter(
        str(node.get("group") or "")
        for node in semantic.get("nodes") or []
        if str(node.get("group") or "")
    )
    node_kinds = Counter(
        str(node.get("kind") or "")
        for node in semantic.get("nodes") or []
        if str(node.get("kind") or "")
    )
    edge_labels = Counter(
        str(edge.get("label") or "")
        for edge in semantic.get("edges") or []
        if str(edge.get("label") or "")
    )
    provider_clusters = [
        {"provider": provider, "nodeCount": count}
        for provider, count in node_groups.most_common(max(limit, 0))
        if provider != "content"
    ]
    dominant_kinds = [
        {"kind": kind, "nodeCount": count}
        for kind, count in node_kinds.most_common(max(limit, 0))
    ]
    dominant_relations = [
        {"label": label, "edgeCount": count}
        for label, count in edge_labels.most_common(max(limit, 0))
    ]
    structural_edge_count = sum(
        count
        for label, count in edge_labels.items()
        if label in {"source", "resource", "resolved_by", "query", "drives"}
    )
    source_node_count = sum(
        count
        for kind, count in node_kinds.items()
        if kind in {"source", "query", "provider"}
    )
    anchors = sorted(
        [dict(item) for item in lineage.get("content") or []],
        key=lambda item: (
            1 if str(item.get("artifactKind") or "") == "evidence" else 0,
            int(item.get("fetchCount") or 0),
            str(item.get("lastFetchedAt") or ""),
            str(item.get("preferredName") or ""),
        ),
        reverse=True,
    )[: max(limit, 0)]
    queries = [
        dict(node)
        for node in semantic.get("nodes") or []
        if str(node.get("kind") or "") == "query"
    ][: max(limit, 0)]
    lead_sources = [
        dict(item) for item in (lineage.get("leadSources") or [])[: max(limit, 0)]
    ]
    return {
        "sessionDir": lineage["sessionDir"],
        "contentDir": lineage["contentDir"],
        "manifestPath": lineage["manifestPath"],
        "contentCount": int(lineage.get("contentCount") or 0),
        "sourceCount": int(lineage.get("sourceCount") or 0),
        "leadSourceCount": int(lineage.get("leadSourceCount") or 0),
        "leadEdgeCount": int(lineage.get("leadEdgeCount") or 0),
        "semanticNodeCount": int(semantic.get("nodeCount") or 0),
        "semanticEdgeCount": int(semantic.get("edgeCount") or 0),
        "discoveryArtifactCount": int(lineage.get("discoveryArtifactCount") or 0),
        "evidenceArtifactCount": int(lineage.get("evidenceArtifactCount") or 0),
        "nextStep": str(lineage.get("nextStep") or semantic.get("nextStep") or ""),
        "providerClusters": provider_clusters,
        "dominantKinds": dominant_kinds,
        "dominantRelations": dominant_relations,
        "materializedAnchors": anchors,
        "querySeeds": queries,
        "bestLeads": lead_sources,
        "sourceHeavy": source_node_count * 2
        >= max(int(semantic.get("nodeCount") or 0), 1),
        "structuralHeavy": structural_edge_count * 2
        >= max(int(semantic.get("edgeCount") or 0), 1),
        "sourceNodeCount": source_node_count,
        "structuralEdgeCount": structural_edge_count,
    }


def _render_analysis_overview_text(payload: dict[str, object]) -> str:
    lines = [
        f"session: {payload['sessionDir']}",
        f"content: {payload['contentDir']}",
        (
            "artifacts: "
            f"{payload['contentCount']} "
            f"(discovery {payload['discoveryArtifactCount']}, "
            f"evidence {payload['evidenceArtifactCount']})"
        ),
        (
            "graph: "
            f"{payload['sourceCount']} sources, "
            f"{payload['leadSourceCount']} lead sources, "
            f"{payload['leadEdgeCount']} lead edges"
        ),
        (
            "semantic: "
            f"{payload['semanticNodeCount']} nodes, "
            f"{payload['semanticEdgeCount']} edges"
        ),
    ]
    shape_parts: list[str] = []
    if bool(payload.get("sourceHeavy")):
        shape_parts.append(
            f"source-heavy ({int(payload['sourceNodeCount'])}/{int(payload['semanticNodeCount'])} source/query/provider nodes)"
        )
    if bool(payload.get("structuralHeavy")):
        shape_parts.append(
            f"structural-edge-heavy ({int(payload['structuralEdgeCount'])}/{int(payload['semanticEdgeCount'])} structural edges)"
        )
    if shape_parts:
        lines.append("shape: " + "; ".join(shape_parts))
    if payload.get("nextStep"):
        lines.append(f"next: {payload['nextStep']}")
    if payload["providerClusters"]:
        lines.append("provider clusters:")
        for cluster in payload["providerClusters"]:
            lines.append(f"  - {cluster['provider']}: {cluster['nodeCount']} nodes")
    if payload["dominantKinds"]:
        lines.append("dominant node kinds:")
        for item in payload["dominantKinds"]:
            lines.append(f"  - {item['kind']}: {item['nodeCount']}")
    if payload["dominantRelations"]:
        lines.append("dominant relations:")
        for item in payload["dominantRelations"]:
            lines.append(f"  - {item['label']}: {item['edgeCount']}")
    if payload["materializedAnchors"]:
        lines.append("materialized anchors:")
        for anchor in payload["materializedAnchors"]:
            providers = ", ".join(str(value) for value in anchor.get("providers") or [])
            visibility = _visibility_summary(anchor)
            lines.append(
                f"  - [{anchor.get('artifactKind') or 'artifact'}] {anchor['preferredName']}"
            )
            if providers:
                lines.append(f"    providers: {providers}")
            if visibility:
                lines.append(f"    visibility: {visibility}")
            lines.append(f"    follow: `{anchor['followCommand']}`")
    if payload["querySeeds"]:
        lines.append("query seeds:")
        for node in payload["querySeeds"]:
            lines.append(f"  - {node['label']}")
    if payload["bestLeads"]:
        lines.append("best leads:")
        for lead in payload["bestLeads"]:
            relation = ", ".join(
                str(value) for value in lead.get("relationKinds") or [] if str(value)
            )
            lines.append(
                f"  - [{'; '.join(_lead_signal_labels(lead, aggregated=True))}] "
                f"{lead['locator']} ({lead['provider']}, {relation or 'lead'})"
            )
            lines.append(f"    follow: `{lead['followCommand']}`")
    lines.append(
        "focus: use `gotta session analyze --focus <locator|keyword> --session <session>` "
        "to inspect one local neighborhood instead of dumping the full graph."
    )
    return "\n".join(lines)


def _render_lineage_overview_text(
    payload: dict[str, object],
    *,
    limit: int,
) -> str:
    lines = [
        f"session: {payload['sessionDir']}",
        f"content: {payload['contentDir']}",
        (
            "artifacts: "
            f"{payload['contentCount']} "
            f"(discovery {payload['discoveryArtifactCount']}, "
            f"evidence {payload['evidenceArtifactCount']})"
        ),
        (
            "graph: "
            f"{payload['sourceCount']} sources, "
            f"{payload['leadSourceCount']} lead sources, "
            f"{payload['leadEdgeCount']} lead edges"
        ),
    ]
    if payload.get("nextStep"):
        lines.append(f"next: {payload['nextStep']}")
    sources = [dict(item) for item in (payload.get("sources") or [])[: max(limit, 0)]]
    if sources:
        lines.append("materialized sources:")
        for source in sources:
            lines.append(
                f"  - {source['locator']} "
                f"({int(source.get('contentCount') or 0)} materializations)"
            )
    if payload["leadSources"]:
        lines.append("best leads:")
        for lead in (payload.get("leadSources") or [])[: max(limit, 0)]:
            relation = ", ".join(
                str(value) for value in lead.get("relationKinds") or [] if str(value)
            )
            lines.append(
                f"  - [{'; '.join(_lead_signal_labels(lead, aggregated=True))}] "
                f"{lead['locator']} ({lead['provider']}, {relation or 'lead'})"
            )
            lines.append(f"    follow: `{lead['followCommand']}`")
    return "\n".join(lines)


def _render_semantic_overview_text(payload: dict[str, object]) -> str:
    lines = [
        f"session: {payload['sessionDir']}",
        f"content: {payload['contentDir']}",
        (
            "artifacts: "
            f"{payload['contentCount']} "
            f"(discovery {payload['discoveryArtifactCount']}, "
            f"evidence {payload['evidenceArtifactCount']})"
        ),
        (
            "semantic: "
            f"{payload['semanticNodeCount']} nodes, "
            f"{payload['semanticEdgeCount']} edges"
        ),
    ]
    shape_parts: list[str] = []
    if bool(payload.get("sourceHeavy")):
        shape_parts.append(
            f"source-heavy ({int(payload['sourceNodeCount'])}/{int(payload['semanticNodeCount'])} source/query/provider nodes)"
        )
    if bool(payload.get("structuralHeavy")):
        shape_parts.append(
            f"structural-edge-heavy ({int(payload['structuralEdgeCount'])}/{int(payload['semanticEdgeCount'])} structural edges)"
        )
    if shape_parts:
        lines.append("shape: " + "; ".join(shape_parts))
    if payload.get("nextStep"):
        lines.append(f"next: {payload['nextStep']}")
    if payload["providerClusters"]:
        lines.append("provider clusters:")
        for cluster in payload["providerClusters"]:
            lines.append(f"  - {cluster['provider']}: {cluster['nodeCount']} nodes")
    if payload["dominantKinds"]:
        lines.append("dominant node kinds:")
        for item in payload["dominantKinds"]:
            lines.append(f"  - {item['kind']}: {item['nodeCount']}")
    if payload["dominantRelations"]:
        lines.append("dominant relations:")
        for item in payload["dominantRelations"]:
            lines.append(f"  - {item['label']}: {item['edgeCount']}")
    if payload["querySeeds"]:
        lines.append("query seeds:")
        for node in payload["querySeeds"]:
            lines.append(f"  - {node['label']}")
    return "\n".join(lines)


def _render_analysis_focus_text(payload: dict[str, object]) -> str:
    lines = [
        f"session: {payload['sessionDir']}",
        f"focus: {payload['focus'] or '(empty)'}",
    ]
    if not payload.get("matched"):
        if payload.get("nextStep"):
            lines.append(f"next: {payload['nextStep']}")
        return "\n".join(lines)
    root = payload["root"]
    lines.append(f"matched: {root['label']} ({root['kind']}, {root['group']})")
    if int(payload.get("matchedCount") or 0) > 1:
        lines.append(
            f"signal: {int(payload['matchedCount'])} anchors matched this focus; "
            "showing the strongest root plus nearby corroborating anchors"
        )
    state_bits = []
    if root.get("artifactKind"):
        state_bits.append(f"artifact_kind={root['artifactKind']}")
    if bool(root.get("materialized")):
        state_bits.append("materialized")
    if bool(root.get("discovered")) and not bool(root.get("materialized")):
        state_bits.append("discovered-only")
    if state_bits:
        lines.append("state: " + ", ".join(state_bits))
    if root.get("followCommand"):
        lines.append(f"follow: `{root['followCommand']}`")
    anchors = payload.get("anchors") or []
    if anchors:
        lines.append("also matched:")
        for anchor in anchors:
            lines.append(f"  - {anchor['label']} ({anchor['kind']}, {anchor['group']})")
            if anchor.get("followCommand"):
                lines.append(f"    follow: `{anchor['followCommand']}`")
    if int(payload.get("suppressedStructuralEdgeCount") or 0) > 0:
        lines.append(
            "signal: "
            f"suppressed {payload['suppressedStructuralEdgeCount']} lower-signal structural edges "
            "to keep this neighborhood readable"
        )
    if payload["neighbors"]:
        lines.append("neighbors:")
        for neighbor in payload["neighbors"]:
            relation = ", ".join(
                str(value) for value in neighbor.get("relations") or [] if str(value)
            )
            lines.append(
                f"  - {neighbor['label']} ({neighbor['kind']}, {neighbor['group']}; {relation or 'adjacent'})"
            )
            bits = []
            if neighbor.get("artifactKind"):
                bits.append(f"artifact_kind={neighbor['artifactKind']}")
            if bool(neighbor.get("materialized")):
                bits.append("materialized")
            if bool(neighbor.get("discovered")) and not bool(
                neighbor.get("materialized")
            ):
                bits.append("discovered-only")
            if bits:
                lines.append("    state: " + ", ".join(bits))
            if neighbor.get("followCommand"):
                lines.append(f"    follow: `{neighbor['followCommand']}`")
    else:
        lines.append("neighbors: none")
    return "\n".join(lines)


def _render_lineage_focus_text(payload: dict[str, object]) -> str:
    lines = [
        f"session: {payload['sessionDir']}",
        f"focus: {payload['focus'] or '(empty)'}",
    ]
    if not payload.get("matched"):
        if payload.get("nextStep"):
            lines.append(f"next: {payload['nextStep']}")
        return "\n".join(lines)
    root = payload["root"]
    lines.append(f"matched: {root['label']} ({root['kind']})")
    if int(payload.get("matchedCount") or 0) > 1:
        lines.append(
            f"signal: {int(payload['matchedCount'])} anchors matched this focus; "
            "showing the strongest root plus nearby corroborating anchors"
        )
    state_bits = []
    if root.get("artifactKind"):
        state_bits.append(f"artifact_kind={root['artifactKind']}")
    if bool(root.get("materialized")):
        state_bits.append("materialized")
    else:
        state_bits.append("discovered-only")
    if state_bits:
        lines.append("state: " + ", ".join(state_bits))
    if root.get("followCommand"):
        lines.append(f"follow: `{root['followCommand']}`")
    anchors = payload.get("anchors") or []
    if anchors:
        lines.append("also matched:")
        for anchor in anchors:
            lines.append(f"  - {anchor['label']} ({anchor['kind']})")
            if anchor.get("followCommand"):
                lines.append(f"    follow: `{anchor['followCommand']}`")
    if payload["neighbors"]:
        lines.append("neighbors:")
        for neighbor in payload["neighbors"]:
            lines.append(
                f"  - {neighbor['label']} ({neighbor['kind']}; {neighbor['relation']})"
            )
            bits = []
            if neighbor.get("artifactKind"):
                bits.append(f"artifact_kind={neighbor['artifactKind']}")
            bits.append(
                "materialized"
                if bool(neighbor.get("materialized"))
                else "discovered-only"
            )
            lines.append("    state: " + ", ".join(bits))
            if neighbor.get("followCommand"):
                lines.append(f"    follow: `{neighbor['followCommand']}`")
    else:
        lines.append("neighbors: none")
    return "\n".join(lines)


def _render_text_bundle(sections: list[tuple[str, str]]) -> str:
    lines: list[str] = []
    for title, body in sections:
        body_text = str(body).strip()
        if not body_text:
            continue
        if lines:
            lines.append("")
        lines.append(f"## {title}")
        lines.append("")
        lines.append(body_text)
    return "\n".join(lines)


def _render_markdown_list(lines: list[str], heading: str, entries: list[str]) -> None:
    if not entries:
        return
    lines.extend([f"## {heading}", ""])
    lines.extend([f"- {entry}" for entry in entries])
    lines.append("")


def _render_lineage_overview_markdown(payload: dict[str, object], *, limit: int) -> str:
    lines = [
        "# gotta session analyze",
        "",
        f"Session: `{payload['sessionDir']}`",
        "",
        (
            f"Artifacts: {payload['contentCount']} total "
            f"(discovery {payload['discoveryArtifactCount']}, "
            f"evidence {payload['evidenceArtifactCount']})"
        ),
        (
            f"Lineage: {payload['sourceCount']} sources, "
            f"{payload['leadSourceCount']} lead sources, "
            f"{payload['leadEdgeCount']} lead edges"
        ),
        "",
    ]
    next_step = str(payload.get("nextStep") or "").strip()
    if next_step:
        lines.extend(["## Synthesis", "", next_step, ""])
    sources = [
        f"{source['locator']} ({int(source.get('contentCount') or 0)} materializations)"
        for source in list(payload.get("sources") or [])[: max(limit, 0)]
    ]
    _render_markdown_list(lines, "Materialized Sources", sources)
    leads_preview = []
    for lead in list(payload.get("leadSources") or [])[: max(limit, 0)]:
        relation = ", ".join(
            str(value) for value in lead.get("relationKinds") or [] if str(value)
        )
        label = (
            f"[{'; '.join(_lead_signal_labels(lead, aggregated=True))}] "
            f"{lead['locator']} ({lead['provider']}, {relation or 'lead'})"
        )
        follow = str(lead.get("followCommand") or "").strip()
        if follow:
            label += f" via `{follow}`"
        leads_preview.append(label)
    _render_markdown_list(lines, "Lead Preview", leads_preview)
    while lines and lines[-1] == "":
        lines.pop()
    return "\n".join(lines) + "\n"


def _render_semantic_overview_markdown(payload: dict[str, object]) -> str:
    lines = [
        "# gotta session analyze",
        "",
        f"Session: `{payload['sessionDir']}`",
        "",
        (
            f"Artifacts: {payload['contentCount']} total "
            f"(discovery {payload['discoveryArtifactCount']}, "
            f"evidence {payload['evidenceArtifactCount']})"
        ),
        (
            f"Semantic: {payload['semanticNodeCount']} nodes, "
            f"{payload['semanticEdgeCount']} edges"
        ),
        "",
    ]
    shape_parts: list[str] = []
    if bool(payload.get("sourceHeavy")):
        shape_parts.append(
            f"source-heavy ({int(payload['sourceNodeCount'])}/{int(payload['semanticNodeCount'])} source/query/provider nodes)"
        )
    if bool(payload.get("structuralHeavy")):
        shape_parts.append(
            f"structural-edge-heavy ({int(payload['structuralEdgeCount'])}/{int(payload['semanticEdgeCount'])} structural edges)"
        )
    if shape_parts:
        lines.extend(["## Shape", "", "; ".join(shape_parts), ""])
    next_step = str(payload.get("nextStep") or "").strip()
    if next_step:
        lines.extend(["## Synthesis", "", next_step, ""])
    provider_clusters = [
        f"{cluster['provider']}: {cluster['nodeCount']} nodes"
        for cluster in list(payload.get("providerClusters") or [])[
            :SUMMARY_BUCKET_LIMIT
        ]
    ]
    _render_markdown_list(lines, "Provider Clusters", provider_clusters)
    dominant_kinds = [
        f"{item['kind']}: {item['nodeCount']}"
        for item in list(payload.get("dominantKinds") or [])[:SUMMARY_BUCKET_LIMIT]
    ]
    _render_markdown_list(lines, "Dominant Node Kinds", dominant_kinds)
    dominant_relations = [
        f"{item['label']}: {item['edgeCount']} edges"
        for item in list(payload.get("dominantRelations") or [])[:SUMMARY_BUCKET_LIMIT]
    ]
    _render_markdown_list(lines, "Dominant Relations", dominant_relations)
    query_seeds = [
        str(node.get("label") or "").strip()
        for node in list(payload.get("querySeeds") or [])[:SUMMARY_BUCKET_LIMIT]
        if str(node.get("label") or "").strip()
    ]
    _render_markdown_list(lines, "Query Seeds", query_seeds)
    while lines and lines[-1] == "":
        lines.pop()
    return "\n".join(lines) + "\n"


def _render_lineage_focus_markdown_section(payload: dict[str, object]) -> list[str]:
    lines: list[str] = ["## Lineage Focus", ""]
    if not payload.get("matched"):
        lines.append("- Match: none")
        next_step = str(payload.get("nextStep") or "").strip()
        if next_step:
            lines.append(f"- Next: {next_step}")
        lines.append("")
        return lines
    root = payload["root"]
    lines.append(f"- Matched: `{root['label']}` ({root['kind']})")
    if int(payload.get("matchedCount") or 0) > 1:
        lines.append(
            f"- Signal: {int(payload['matchedCount'])} anchors matched this focus; showing the strongest root plus nearby corroborating anchors"
        )
    state_bits = []
    if root.get("artifactKind"):
        state_bits.append(f"artifact_kind={root['artifactKind']}")
    state_bits.append(
        "materialized" if bool(root.get("materialized")) else "discovered-only"
    )
    if state_bits:
        lines.append(f"- State: {', '.join(state_bits)}")
    follow = str(root.get("followCommand") or "").strip()
    if follow:
        lines.append(f"- Follow: `{follow}`")
    anchors = list(payload.get("anchors") or [])[:ANALYZE_FOCUS_MATCH_PREVIEW_LIMIT]
    if anchors:
        lines.extend(["", "### Also Matched", ""])
        for anchor in anchors:
            entry = f"`{anchor['label']}` ({anchor['kind']})"
            follow = str(anchor.get("followCommand") or "").strip()
            if follow:
                entry += f" via `{follow}`"
            lines.append(f"- {entry}")
        hidden_anchors = max(
            int(len(payload.get("anchors") or [])) - len(anchors),
            0,
        )
        if hidden_anchors > 0:
            lines.append(f"- ... {hidden_anchors} additional matched anchors hidden")
    neighbors = list(payload.get("neighbors") or [])[
        :ANALYZE_FOCUS_NEIGHBOR_PREVIEW_LIMIT
    ]
    lines.extend(["", "### Neighbors", ""])
    if neighbors:
        for neighbor in neighbors:
            bits = []
            if neighbor.get("artifactKind"):
                bits.append(f"artifact_kind={neighbor['artifactKind']}")
            bits.append(
                "materialized"
                if bool(neighbor.get("materialized"))
                else "discovered-only"
            )
            entry = (
                f"`{neighbor['label']}` ({neighbor['kind']}; {neighbor['relation']}; "
                f"{', '.join(bits)})"
            )
            follow = str(neighbor.get("followCommand") or "").strip()
            if follow:
                entry += f" via `{follow}`"
            lines.append(f"- {entry}")
        hidden_neighbors = max(
            int(len(payload.get("neighbors") or [])) - len(neighbors),
            0,
        )
        if hidden_neighbors > 0:
            lines.append(f"- ... {hidden_neighbors} additional neighbors hidden")
    else:
        lines.append("- None")
    lines.append("")
    return lines


def _render_semantic_focus_markdown_section(payload: dict[str, object]) -> list[str]:
    lines: list[str] = ["## Semantic Focus", ""]
    if not payload.get("matched"):
        lines.append("- Match: none")
        next_step = str(payload.get("nextStep") or "").strip()
        if next_step:
            lines.append(f"- Next: {next_step}")
        lines.append("")
        return lines
    root = payload["root"]
    lines.append(f"- Matched: `{root['label']}` ({root['kind']}, {root['group']})")
    if int(payload.get("matchedCount") or 0) > 1:
        lines.append(
            f"- Signal: {int(payload['matchedCount'])} anchors matched this focus; showing the strongest root plus nearby corroborating anchors"
        )
    state_bits = []
    if root.get("artifactKind"):
        state_bits.append(f"artifact_kind={root['artifactKind']}")
    if bool(root.get("materialized")):
        state_bits.append("materialized")
    if bool(root.get("discovered")) and not bool(root.get("materialized")):
        state_bits.append("discovered-only")
    if state_bits:
        lines.append(f"- State: {', '.join(state_bits)}")
    follow = str(root.get("followCommand") or "").strip()
    if follow:
        lines.append(f"- Follow: `{follow}`")
    anchors = list(payload.get("anchors") or [])[:ANALYZE_FOCUS_MATCH_PREVIEW_LIMIT]
    if anchors:
        lines.extend(["", "### Also Matched", ""])
        for anchor in anchors:
            entry = f"`{anchor['label']}` ({anchor['kind']}, {anchor['group']})"
            follow = str(anchor.get("followCommand") or "").strip()
            if follow:
                entry += f" via `{follow}`"
            lines.append(f"- {entry}")
        hidden_anchors = max(
            int(len(payload.get("anchors") or [])) - len(anchors),
            0,
        )
        if hidden_anchors > 0:
            lines.append(f"- ... {hidden_anchors} additional matched anchors hidden")
    suppressed = int(payload.get("suppressedStructuralEdgeCount") or 0)
    if suppressed > 0:
        lines.append("")
        lines.append(
            f"- Signal: suppressed {suppressed} lower-signal structural edges to keep this neighborhood readable"
        )
    neighbors = list(payload.get("neighbors") or [])[
        :ANALYZE_FOCUS_NEIGHBOR_PREVIEW_LIMIT
    ]
    lines.extend(["", "### Neighbors", ""])
    if neighbors:
        for neighbor in neighbors:
            relation = ", ".join(
                str(value) for value in neighbor.get("relations") or [] if str(value)
            )
            bits = []
            if neighbor.get("artifactKind"):
                bits.append(f"artifact_kind={neighbor['artifactKind']}")
            if bool(neighbor.get("materialized")):
                bits.append("materialized")
            if bool(neighbor.get("discovered")) and not bool(
                neighbor.get("materialized")
            ):
                bits.append("discovered-only")
            entry = (
                f"`{neighbor['label']}` ({neighbor['kind']}, {neighbor['group']}; "
                f"{relation or 'adjacent'}"
            )
            if bits:
                entry += f"; {', '.join(bits)}"
            entry += ")"
            follow = str(neighbor.get("followCommand") or "").strip()
            if follow:
                entry += f" via `{follow}`"
            lines.append(f"- {entry}")
        hidden_neighbors = max(
            int(len(payload.get("neighbors") or [])) - len(neighbors),
            0,
        )
        if hidden_neighbors > 0:
            lines.append(f"- ... {hidden_neighbors} additional neighbors hidden")
    else:
        lines.append("- None")
    lines.append("")
    return lines


def _render_combined_focus_markdown(
    *, lineage: dict[str, object], semantic: dict[str, object]
) -> str:
    focus = str(lineage.get("focus") or semantic.get("focus") or "").strip()
    lines = [
        "# gotta session analyze",
        "",
        f"Session: `{lineage['sessionDir']}`",
        "",
        f"Focus: `{focus or '(empty)'}`",
        "",
    ]
    lines.extend(_render_lineage_focus_markdown_section(lineage))
    lines.extend(_render_semantic_focus_markdown_section(semantic))
    while lines and lines[-1] == "":
        lines.pop()
    return "\n".join(lines) + "\n"


def _render_single_focus_markdown(
    *,
    session_dir: str,
    focus: str,
    section_lines: list[str],
) -> str:
    lines = [
        "# gotta session analyze",
        "",
        f"Session: `{session_dir}`",
        "",
        f"Focus: `{focus or '(empty)'}`",
        "",
    ]
    lines.extend(section_lines)
    while lines and lines[-1] == "":
        lines.pop()
    return "\n".join(lines) + "\n"


def _render_analysis_overview_markdown(
    overview: dict[str, object],
    *,
    lineage: dict[str, object],
    limit: int,
) -> str:
    lines = [
        "# gotta session analyze",
        "",
        f"Session: `{overview['sessionDir']}`",
        "",
        (
            f"Artifacts: {overview['contentCount']} total "
            f"(discovery {overview['discoveryArtifactCount']}, "
            f"evidence {overview['evidenceArtifactCount']})"
        ),
        (
            f"Lineage: {overview['sourceCount']} sources, "
            f"{overview['leadSourceCount']} lead sources, "
            f"{overview['leadEdgeCount']} lead edges"
        ),
        (
            f"Semantic: {overview['semanticNodeCount']} nodes, "
            f"{overview['semanticEdgeCount']} edges"
        ),
        "",
    ]
    next_step = str(overview.get("nextStep") or "").strip()
    if next_step:
        lines.extend(["## Synthesis", "", next_step, ""])
    provider_clusters = [
        f"{item['provider']}: {item['nodeCount']} nodes"
        for item in list(overview.get("providerClusters") or [])[:SUMMARY_BUCKET_LIMIT]
    ]
    _render_markdown_list(lines, "Provider Clusters", provider_clusters)
    dominant_relations = [
        f"{item['label']}: {item['edgeCount']} edges"
        for item in list(overview.get("dominantRelations") or [])[:SUMMARY_BUCKET_LIMIT]
    ]
    _render_markdown_list(lines, "Dominant Relations", dominant_relations)
    anchors = []
    for anchor in list(overview.get("materializedAnchors") or [])[
        : max(limit, ANALYZE_ANCHOR_PREVIEW_LIMIT)
    ]:
        follow = str(anchor.get("followCommand") or "").strip()
        if follow:
            anchors.append(
                f"[{anchor.get('artifactKind') or 'artifact'}] {anchor['preferredName']} via `{follow}`"
            )
        else:
            anchors.append(
                f"[{anchor.get('artifactKind') or 'artifact'}] {anchor['preferredName']}"
            )
    _render_markdown_list(lines, "Anchor Shortlist", anchors)
    lineage_preview = [
        f"{item['locator']} ({int(item.get('contentCount') or 0)} materializations)"
        for item in list(lineage.get("sources") or [])[
            : max(limit, ANALYZE_ANCHOR_PREVIEW_LIMIT)
        ]
    ]
    _render_markdown_list(lines, "Lineage Preview", lineage_preview)
    semantic_preview = [
        f"{item['label']} ({item['kind']})"
        for item in list(overview.get("querySeeds") or [])[
            : max(limit, ANALYZE_ANCHOR_PREVIEW_LIMIT)
        ]
    ]
    _render_markdown_list(lines, "Semantic Preview", semantic_preview)
    leads_preview = []
    for lead in list(overview.get("bestLeads") or [])[
        : max(limit, ANALYZE_ANCHOR_PREVIEW_LIMIT)
    ]:
        follow = str(lead.get("followCommand") or "").strip()
        relation = ", ".join(
            str(value) for value in lead.get("relationKinds") or [] if str(value)
        )
        label = f"{lead['locator']} ({lead['provider']}, {relation or 'lead'})"
        if follow:
            label += f" via `{follow}`"
        leads_preview.append(label)
    _render_markdown_list(lines, "Lead Preview", leads_preview)
    while lines and lines[-1] == "":
        lines.pop()
    return "\n".join(lines) + "\n"


def _render_markdown_bundle(
    sections: list[tuple[str, str]],
    *,
    title: str = "gotta session analyze",
) -> str:
    lines = [f"# {title}", ""]
    for index, (section_title, body) in enumerate(sections):
        graph_text = str(body).rstrip()
        if index:
            lines.append("")
        lines.append(f"## {section_title}")
        lines.append("")
        lines.append("```mermaid")
        lines.append(graph_text)
        lines.append("```")
    return "\n".join(lines)


def _combined_analysis_payload(
    *,
    focus: str,
    lineage: dict[str, object],
    semantic: dict[str, object],
) -> dict[str, object]:
    return {
        "mode": "all",
        "focus": focus,
        "lineage": lineage,
        "semantic": semantic,
    }


def cmd_analyze(args: argparse.Namespace) -> int:
    dirs = _session_dirs_for_read(args)
    _require_started_session(dirs)
    session_ref = _explicit_session_ref(args)
    payload = _analysis_payload(dirs, session_ref=session_ref)
    mermaid = _render_analysis_mermaid(payload)
    semantic_payload = _semantic_payload(dirs, session_ref=session_ref)
    semantic_mermaid = _render_semantic_mermaid(semantic_payload)
    focus_query = str(getattr(args, "focus", "") or "").strip()
    focus_limit = max(int(getattr(args, "limit", 8) or 0), 0)
    overview = _analysis_overview_payload(
        payload,
        semantic_payload,
        limit=focus_limit,
    )
    lineage_focus_payload = None
    semantic_focus_payload = None
    if focus_query:
        focus_scan_payload = _scan_payload(
            dirs,
            query=focus_query,
            limit=max(focus_limit * 2, 12),
            include_all=True,
            session_ref=session_ref,
        )
        lineage_focus_payload = _lineage_focus_payload(
            payload,
            focus=focus_query,
            limit=focus_limit,
            scan_payload=focus_scan_payload,
        )
        semantic_focus_payload = _semantic_focus_payload(
            payload,
            semantic_payload,
            focus=focus_query,
            limit=focus_limit,
            scan_payload=focus_scan_payload,
        )

    if args.output == "text":
        if args.mode == "lineage":
            print(
                _render_lineage_focus_text(lineage_focus_payload)
                if lineage_focus_payload is not None
                else _render_lineage_overview_text(payload, limit=focus_limit)
            )
        elif args.mode == "semantic":
            print(
                _render_analysis_focus_text(semantic_focus_payload)
                if semantic_focus_payload is not None
                else _render_semantic_overview_text(overview)
            )
        elif focus_query:
            print(
                _render_text_bundle(
                    [
                        ("Lineage", _render_lineage_focus_text(lineage_focus_payload)),
                        (
                            "Semantic",
                            _render_analysis_focus_text(semantic_focus_payload),
                        ),
                    ]
                )
            )
        else:
            print(_render_analysis_overview_text(overview))
    elif args.output == "json":
        if args.mode == "lineage":
            print(
                json.dumps(
                    lineage_focus_payload
                    if lineage_focus_payload is not None
                    else payload,
                    indent=2,
                    sort_keys=True,
                )
            )
        elif args.mode == "semantic":
            print(
                json.dumps(
                    semantic_focus_payload
                    if semantic_focus_payload is not None
                    else semantic_payload,
                    indent=2,
                    sort_keys=True,
                )
            )
        else:
            print(
                json.dumps(
                    _combined_analysis_payload(
                        focus=focus_query,
                        lineage=(
                            lineage_focus_payload
                            if lineage_focus_payload is not None
                            else payload
                        ),
                        semantic=(
                            semantic_focus_payload
                            if semantic_focus_payload is not None
                            else semantic_payload
                        ),
                    ),
                    indent=2,
                    sort_keys=True,
                )
            )
    elif args.output == "markdown":
        if args.mode == "lineage":
            print(
                _render_single_focus_markdown(
                    session_dir=str(lineage_focus_payload["sessionDir"]),
                    focus=str(lineage_focus_payload.get("focus") or ""),
                    section_lines=_render_lineage_focus_markdown_section(
                        lineage_focus_payload
                    ),
                )
                if lineage_focus_payload is not None
                else _render_lineage_overview_markdown(payload, limit=focus_limit),
                end="",
            )
        elif args.mode == "semantic":
            print(
                _render_single_focus_markdown(
                    session_dir=str(semantic_focus_payload["sessionDir"]),
                    focus=str(semantic_focus_payload.get("focus") or ""),
                    section_lines=_render_semantic_focus_markdown_section(
                        semantic_focus_payload
                    ),
                )
                if semantic_focus_payload is not None
                else _render_semantic_overview_markdown(overview),
                end="",
            )
        elif focus_query:
            print(
                _render_combined_focus_markdown(
                    lineage=lineage_focus_payload,
                    semantic=semantic_focus_payload,
                ),
                end="",
            )
        else:
            print(
                _render_analysis_overview_markdown(
                    overview,
                    lineage=payload,
                    limit=focus_limit,
                ),
                end="",
            )
    else:
        if args.mode == "lineage":
            print(
                _render_analysis_mermaid(lineage_focus_payload)
                if lineage_focus_payload is not None
                else mermaid
            )
        else:
            print(
                _render_semantic_mermaid(semantic_focus_payload)
                if semantic_focus_payload is not None
                else semantic_mermaid
            )
    return 0


analysis_payload = _analysis_payload
semantic_payload = _semantic_payload
render_analysis_mermaid = _render_analysis_mermaid
render_semantic_mermaid = _render_semantic_mermaid
combined_analysis_payload = _combined_analysis_payload
analysis_overview_payload = _analysis_overview_payload
lineage_focus_payload = _lineage_focus_payload
semantic_focus_payload = _semantic_focus_payload
